"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import DetectorViewerShowAsTypes, DetectorViewerShowDataTypes
from zospy.api._ZOSAPI.Analysis.Settings import IAS_, DetectorViewerScaleTypes

__all__ = ("IAS_DetectorViewer", "IAS_NSCSingleRayTrace")

class IAS_DetectorViewer(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Detector(self) -> IAS_Detector: ...
    @property
    def ShowAs(self) -> DetectorViewerShowAsTypes: ...
    @ShowAs.setter
    def ShowAs(self, value: DetectorViewerShowAsTypes) -> None: ...
    @property
    def Scale(self) -> DetectorViewerScaleTypes: ...
    @Scale.setter
    def Scale(self, value: DetectorViewerScaleTypes) -> None: ...
    @property
    def DataType(self) -> DetectorViewerShowDataTypes: ...
    @DataType.setter
    def DataType(self, value: DetectorViewerShowDataTypes) -> None: ...
    @property
    def SymbolType(self) -> int: ...
    @SymbolType.setter
    def SymbolType(self, value: int) -> None: ...
    @property
    def Configuration(self) -> int: ...
    @Configuration.setter
    def Configuration(self, value: int) -> None: ...
    @property
    def ExtraProperty(self) -> int: ...
    @ExtraProperty.setter
    def ExtraProperty(self, value: int) -> None: ...
    @property
    def Smoothing(self) -> int: ...
    @Smoothing.setter
    def Smoothing(self, value: int) -> None: ...
    @property
    def RowCol(self) -> int: ...
    @RowCol.setter
    def RowCol(self, value: int) -> None: ...
    @property
    def Zplane(self) -> int: ...
    @Zplane.setter
    def Zplane(self, value: int) -> None: ...
    @property
    def Contrast(self) -> UInt16: ...
    @Contrast.setter
    def Contrast(self, value: UInt16) -> None: ...
    @property
    def AngleList(self) -> list[UInt16]: ...
    @AngleList.setter
    def AngleList(self, value: list[UInt16]) -> None: ...
    @property
    def PlotScaleMinimum(self) -> float: ...
    @PlotScaleMinimum.setter
    def PlotScaleMinimum(self, value: float) -> None: ...
    @property
    def PlotScaleMaximum(self) -> float: ...
    @PlotScaleMaximum.setter
    def PlotScaleMaximum(self, value: float) -> None: ...
    @property
    def RayDatabaseFilename(self) -> str: ...
    @RayDatabaseFilename.setter
    def RayDatabaseFilename(self, value: str) -> None: ...
    @property
    def Filter(self) -> str: ...
    @Filter.setter
    def Filter(self, value: str) -> None: ...
    @property
    def OutFile(self) -> str: ...
    @OutFile.setter
    def OutFile(self, value: str) -> None: ...
    @property
    def SuppressFrame(self) -> bool: ...
    @SuppressFrame.setter
    def SuppressFrame(self, value: bool) -> None: ...
    @property
    def NumberOfNonSequentialSurfaces(self) -> int: ...
    @property
    def NumberOfDetectorsOnSurface(self) -> int: ...
    @property
    def NumberOfShowAsTypes(self) -> int: ...
    @property
    def MaxSpatialFrequency(self) -> float: ...
    @MaxSpatialFrequency.setter
    def MaxSpatialFrequency(self, value: float) -> None: ...
    def GetDetectorName(self, index: int) -> str: ...
    def GetDetectorObjectNumber(self, index: int) -> int: ...
    def GetNonSequentialSurfaceName(self, index: int) -> str: ...
    def GetNonSequentialSurfaceNumber(self, index: int) -> int: ...
    def GetShowAsTypeNameAt(self, idx: int) -> str: ...
    def IsValidDetector(self, N: int) -> bool: ...

class IAS_NSCSingleRayTrace(IAS_):
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def RaySourceX(self) -> float: ...
    @RaySourceX.setter
    def RaySourceX(self, value: float) -> None: ...
    @property
    def RaySourceY(self) -> float: ...
    @RaySourceY.setter
    def RaySourceY(self, value: float) -> None: ...
    @property
    def RaySourceZ(self) -> float: ...
    @RaySourceZ.setter
    def RaySourceZ(self, value: float) -> None: ...
    @property
    def RaySourceL(self) -> float: ...
    @RaySourceL.setter
    def RaySourceL(self, value: float) -> None: ...
    @property
    def RaySourceM(self) -> float: ...
    @RaySourceM.setter
    def RaySourceM(self, value: float) -> None: ...
    @property
    def RaySourceN(self) -> float: ...
    @RaySourceN.setter
    def RaySourceN(self, value: float) -> None: ...
    @property
    def RefObject(self) -> int: ...
    @RefObject.setter
    def RefObject(self, value: int) -> None: ...
    @property
    def SplitNSCRays(self) -> bool: ...
    @SplitNSCRays.setter
    def SplitNSCRays(self, value: bool) -> None: ...
    @property
    def ScatterNSCRays(self) -> bool: ...
    @ScatterNSCRays.setter
    def ScatterNSCRays(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def ShowXYZ(self) -> bool: ...
    @ShowXYZ.setter
    def ShowXYZ(self, value: bool) -> None: ...
    @property
    def ShowLMN(self) -> bool: ...
    @ShowLMN.setter
    def ShowLMN(self, value: bool) -> None: ...
    @property
    def ShowPath(self) -> bool: ...
    @ShowPath.setter
    def ShowPath(self, value: bool) -> None: ...
    @property
    def ShowNormal(self) -> bool: ...
    @ShowNormal.setter
    def ShowNormal(self, value: bool) -> None: ...
    @property
    def ShowExyz(self) -> bool: ...
    @ShowExyz.setter
    def ShowExyz(self, value: bool) -> None: ...
    @property
    def ExpandIntoBranches(self) -> bool: ...
    @ExpandIntoBranches.setter
    def ExpandIntoBranches(self, value: bool) -> None: ...
    @property
    def DecimalPrecision(self) -> int: ...
    @DecimalPrecision.setter
    def DecimalPrecision(self, value: int) -> None: ...
    @property
    def SaveRaysFile(self) -> str: ...
    @SaveRaysFile.setter
    def SaveRaysFile(self, value: str) -> None: ...



"""Base classes for system viewers.

System viewers are slightly different from other analyses: the analyses themselves are not configurable,
but there is a separate tool for each viewer. This module provides the base classes for interacting with these tools.
"""

from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from dataclasses import fields
from typing import TYPE_CHECKING, Annotated, Generic, Literal, Optional, get_args
from warnings import warn

import numpy as np
from pydantic.fields import Field, FieldInfo
from System import Array

from zospy.analyses.base import AnalysisData, AnalysisResult, AnalysisSettings, BaseAnalysisWrapper, OnComplete
from zospy.utils.pyutils import abspath

if TYPE_CHECKING:
    from os import PathLike
    from pathlib import Path

    from zospy.api import _ZOSAPI
    from zospy.zpcore import OpticStudioSystem


__all__ = ("SystemViewerWrapper", "ImageSize")


class SystemViewerWrapper(BaseAnalysisWrapper[Optional[np.ndarray], AnalysisSettings], ABC, Generic[AnalysisSettings]):
    """Base class for SystemViewer analyses."""

    ALLOWED_IMAGE_EXTENSIONS: tuple[str, ...] = ("bmp", "jpeg", "png")

    def __init__(self, settings_kws: dict[str, any]):
        super().__init__(settings_kws=settings_kws)

        self._image_output_file = None

    def __init_subclass__(cls, **kwargs):
        """Determine the settings type of the viewer."""
        if cls._settings_type is AnalysisSettings:
            if hasattr(cls, "__orig_bases__"):
                base = cls.__orig_bases__[0]
                cls._settings_type: type[AnalysisSettings] = get_args(base)[0]
            else:
                cls._settings_type = type(None)  # TODO: Replace with NoneType when dropping support for Python 3.9

        super().__init_subclass__(**kwargs)

    @property
    def image_output_file(self) -> str | Path | None:
        """Path to the image output file."""
        return self._image_output_file

    def _warn_ignored_settings(self) -> None:
        """Check if unsupported parameters are specified and warn the user.

        For OpticStudio versions below 24R1, compare the values of a dictionary with the default values of a function,
        and warn if any are different.
        """
        changed_parameters = []

        for field in fields(type(self.settings)):
            default = field.default.default if isinstance(field.default, FieldInfo) else field.default

            if getattr(self.settings, field.name) != default:
                changed_parameters.append(field.name)

        if len(changed_parameters) > 0:
            warn(
                f"Some parameters were specified but ignored, because viewer exports are only supported from OpticStudio"
                f"24R1: {', '.join(changed_parameters)}"
            )

    def _validate_path(self, path: PathLike | str) -> str:
        str_path = abspath(path, check_directory_only=True)

        if str_path.split(".")[-1] in self.ALLOWED_IMAGE_EXTENSIONS:
            return str_path

        raise ValueError(f"Image file must have one of the following extensions: {self.ALLOWED_IMAGE_EXTENSIONS}")

    def _validate_wavelength(self, wavelength: int | str) -> int:
        if isinstance(wavelength, str):
            if wavelength == "All":
                return -1

            raise ValueError("wavelength must be an integer or 'All'.")

        if wavelength < -1 or wavelength == 0 or wavelength > self.oss.SystemData.Wavelengths.NumberOfWavelengths:
            raise ValueError("wavelength must be -1 or between 1 and the number of wavelengths.")

        if not isinstance(wavelength, int):
            raise TypeError("wavelength must be an integer or 'All'.")

        return wavelength

    def _validate_field(self, field: int | str) -> int:
        if isinstance(field, str):
            if field == "All":
                return -1

            raise ValueError("field must be an integer or 'All'.")

        if field < -1 or field == 0 or field > self.oss.SystemData.Fields.NumberOfFields:
            raise ValueError("field must be -1 or between 1 and the number of fields.")

        if not isinstance(field, int):
            raise TypeError("field must be an integer or 'All'.")

        return field

    def _validate_end_surface(self, start_surface, end_surface: int):
        if end_surface != -1 and end_surface <= start_surface or end_surface > self.oss.LDE.NumberOfSurfaces - 1:
            raise ValueError(
                "end_surface must be -1 or greater than start_surface and less than the number of surfaces."
            )

        if end_surface == -1:
            return self.oss.LDE.NumberOfSurfaces - 1

        return end_surface

    def _close_current_tool(self) -> None:
        """Close the current tool in OpticStudio."""
        if self.oss.Tools.CurrentTool is not None:
            self.oss.Tools.CurrentTool.Close()

    @staticmethod
    def _get_image_data(image_data: _ZOSAPI.Tools.Layouts.IImageExportData | None) -> np.ndarray | None:
        if image_data is None:
            return image_data

        image_size = image_data.Width * image_data.Height

        # In-place updating arrays works only with dotnet arrays
        r_values = Array[int](image_size)  # [0] * image_size
        g_values = Array[int](image_size)  # [0] * image_size
        b_values = Array[int](image_size)  # [0] * image_size

        image_data.FillValues(image_data.Width * image_data.Height, r_values, g_values, b_values)

        return np.stack((r_values, g_values, b_values), axis=-1).reshape(image_data.Height, image_data.Width, 3)

    @abstractmethod
    def configure_layout_tool(
        self,
    ) -> (
        _ZOSAPI.Tools.Layouts.ICrossSectionExport
        | _ZOSAPI.Tools.Layouts.I3DViewerExport
        | _ZOSAPI.Tools.Layouts.IShadedModelExport
        | _ZOSAPI.Tools.Layouts.INSC3DLayoutExport
        | _ZOSAPI.Tools.Layouts.INSCShadedModelExport
    ):
        """Configure the layout tool for the analysis."""

    def run_analysis(self) -> np.ndarray | None:
        """Run the layout tool."""
        layout_tool = self.configure_layout_tool()

        if self.image_output_file is not None:
            layout_tool.SaveImageAsFile = True
            layout_tool.OutputFileName = self.image_output_file
        else:
            layout_tool.SaveImageAsFile = False

        layout_tool.RunAndWaitForCompletion()

        if not layout_tool.Succeeded:
            raise RuntimeError("The system viewer export tool failed to run.")

        image_data = self._get_image_data(layout_tool.ImageExportData) if self.image_output_file is None else None

        layout_tool.Close()

        return image_data

    def run(
        self,
        oss: OpticStudioSystem,
        image_output_file: str | Path | None = None,
        oncomplete: OnComplete | Literal["Close", "Release", "Sustain"] = "Close",
    ) -> AnalysisData:
        """Run the analysis.

        Parameters
        ----------
        **kwargs
        """
        if image_output_file is not None:
            self._image_output_file = self._validate_path(image_output_file)

        self._oss = weakref.proxy(oss)
        self._create_analysis(settings_first=False)

        image_data = None

        if self.oss.ZOS.version >= (24, 1, 0):
            self._close_current_tool()
            image_data = self.run_analysis()
        else:
            self._warn_ignored_settings()

        result = AnalysisResult(
            image_data,
            settings=self.settings,
            metadata=self.analysis.metadata,
            header=self.analysis.header_data,
            messages=self.analysis.messages,
        )

        self._complete(OnComplete(oncomplete))

        return result


_UInt = Annotated[int, Field(gt=0)]
ImageSize = tuple[_UInt, _UInt]



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import SampleSizes
from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = (
    "EncircledEnergyTypes",
    "ExtendedSourceTypes",
    "GeometricLineEdgeSpreadTypes",
    "IAS_DiffractionEncircledEnergy",
    "IAS_ExtendedSourceEncircledEnergy",
    "IAS_GeometricEncircledEnergy",
    "IAS_GeometricLineEdgeSpread",
    "ReferToTypes",
)

class EncircledEnergyTypes:
    Encircled = 1
    X_Only = 2
    Y_Only = 3
    Ensquared = 4

class ExtendedSourceTypes:
    Encircled = 1
    X_Only = 2
    Y_Only = 3
    Ensquared = 4
    X_Distrib = 5
    Y_Distrib = 6

class GeometricLineEdgeSpreadTypes:
    LineEdge = 1
    Line = 2
    Edge = 3

class IAS_DiffractionEncircledEnergy(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def HuygensSample(self) -> SampleSizes: ...
    @HuygensSample.setter
    def HuygensSample(self, value: SampleSizes) -> None: ...
    @property
    def SampleSize(self) -> SampleSizes: ...
    @SampleSize.setter
    def SampleSize(self, value: SampleSizes) -> None: ...
    @property
    def Type(self) -> EncircledEnergyTypes: ...
    @Type.setter
    def Type(self, value: EncircledEnergyTypes) -> None: ...
    @property
    def ReferTo(self) -> ReferToTypes: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferToTypes) -> None: ...
    @property
    def ScatterRays(self) -> bool: ...
    @ScatterRays.setter
    def ScatterRays(self, value: bool) -> None: ...
    @property
    def ShowDiffractionLimit(self) -> bool: ...
    @ShowDiffractionLimit.setter
    def ShowDiffractionLimit(self, value: bool) -> None: ...
    @property
    def UseDashes(self) -> bool: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @property
    def UseHuygensPSF(self) -> bool: ...
    @UseHuygensPSF.setter
    def UseHuygensPSF(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def HuygensDelta(self) -> float: ...
    @HuygensDelta.setter
    def HuygensDelta(self, value: float) -> None: ...
    @property
    def RadiusMaximum(self) -> float: ...
    @RadiusMaximum.setter
    def RadiusMaximum(self, value: float) -> None: ...

class IAS_ExtendedSourceEncircledEnergy(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def Type(self) -> ExtendedSourceTypes: ...
    @Type.setter
    def Type(self, value: ExtendedSourceTypes) -> None: ...
    @property
    def ReferTo(self) -> ReferToTypes: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferToTypes) -> None: ...
    @property
    def RaysX1000(self) -> int: ...
    @RaysX1000.setter
    def RaysX1000(self, value: int) -> None: ...
    @property
    def MultiplyByDiffractionLimit(self) -> bool: ...
    @MultiplyByDiffractionLimit.setter
    def MultiplyByDiffractionLimit(self, value: bool) -> None: ...
    @property
    def RemoveVignettingFactors(self) -> bool: ...
    @RemoveVignettingFactors.setter
    def RemoveVignettingFactors(self, value: bool) -> None: ...
    @property
    def UseDashes(self) -> bool: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def MaximumDistance(self) -> float: ...
    @MaximumDistance.setter
    def MaximumDistance(self, value: float) -> None: ...
    @property
    def FieldSize(self) -> float: ...
    @FieldSize.setter
    def FieldSize(self, value: float) -> None: ...
    @property
    def Rotation(self) -> float: ...
    @Rotation.setter
    def Rotation(self, value: float) -> None: ...
    @property
    def ImageName(self) -> str: ...
    @ImageName.setter
    def ImageName(self, value: str) -> None: ...

class IAS_GeometricEncircledEnergy(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def HuygensSample(self) -> SampleSizes: ...
    @HuygensSample.setter
    def HuygensSample(self, value: SampleSizes) -> None: ...
    @property
    def SampleSize(self) -> SampleSizes: ...
    @SampleSize.setter
    def SampleSize(self, value: SampleSizes) -> None: ...
    @property
    def Type(self) -> EncircledEnergyTypes: ...
    @Type.setter
    def Type(self, value: EncircledEnergyTypes) -> None: ...
    @property
    def ReferTo(self) -> ReferToTypes: ...
    @ReferTo.setter
    def ReferTo(self, value: ReferToTypes) -> None: ...
    @property
    def ScatterRays(self) -> bool: ...
    @ScatterRays.setter
    def ScatterRays(self, value: bool) -> None: ...
    @property
    def ShowDiffractionLimit(self) -> bool: ...
    @ShowDiffractionLimit.setter
    def ShowDiffractionLimit(self, value: bool) -> None: ...
    @property
    def UseDashes(self) -> bool: ...
    @UseDashes.setter
    def UseDashes(self, value: bool) -> None: ...
    @property
    def UseHuygensPSF(self) -> bool: ...
    @UseHuygensPSF.setter
    def UseHuygensPSF(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def HuygensDelta(self) -> float: ...
    @HuygensDelta.setter
    def HuygensDelta(self, value: float) -> None: ...
    @property
    def RadiusMaximum(self) -> float: ...
    @RadiusMaximum.setter
    def RadiusMaximum(self, value: float) -> None: ...

class IAS_GeometricLineEdgeSpread(IAS_):
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def SampleSize(self) -> SampleSizes: ...
    @SampleSize.setter
    def SampleSize(self, value: SampleSizes) -> None: ...
    @property
    def Type(self) -> GeometricLineEdgeSpreadTypes: ...
    @Type.setter
    def Type(self, value: GeometricLineEdgeSpreadTypes) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def RadiusMaximum(self) -> float: ...
    @RadiusMaximum.setter
    def RadiusMaximum(self, value: float) -> None: ...

class ReferToTypes:
    ChiefRay = 0
    Centroid = 1
    Vertex = 2



"""Physical Optics Propagation analysis."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, Union
from warnings import warn

from pandas import DataFrame
from pydantic import Field, model_validator

from zospy.analyses.base import BaseAnalysisWrapper, new_analysis
from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import WavelengthNumber, ZOSAPIConstant  # noqa: TCH001
from zospy.api import constants
from zospy.utils.zputils import standardize_sampling

if TYPE_CHECKING:
    from zospy.zpcore import OpticStudioSystem

__all__ = (
    "PhysicalOpticsPropagation",
    "PhysicalOpticsPropagationSettings",
    "create_beam_parameter_dict",
    "create_fiber_parameter_dict",
)


@analysis_settings
class PhysicalOpticsPropagationSettings:
    """Settings for Physical Optics Propagation analysis.

    See the OpticStudio documentation for a more in depth description of most parameters.

    Attributes
    ----------
    wavelength : str | int
        The wavelength to use in the analysis. Either 'All' or an integer specifying the wavelength number. Defaults to
        'All'.
    field : int
        The field number of the chief ray used to align the beam for the analysis. Should be an integer specifying the
        field number. Defaults to 1.
    start_surface : str | int
        The surface that is to be analyzed. Either 'Ent. Pupil' or an integer specifying the surface number. Defaults to
         1.
    end_surface : str | int
        The surface that is to be analyzed. Either 'Image' or an integer specifying the surface number. Defaults to
        'Image'.
    surface_to_beam : float
        Distance from starting surface to starting beam position. Defaults to 0.0.
    use_polarization : bool
        Use polarization. Defaults to False.
    separate_xy : bool
        Separate propagation in X Y direction. Defaults to False.
    use_disk_storage : bool
        Use disk storage. Defaults to False.
    beam_type : constants.Analysis.PhysicalOptics.POPBeamTypes | str
        The beam type. Defaults to 'GaussianWaist'.
    beam_file:
        The beam file used for the beam definition. Should include the .ZBF extension. The beam files have to be located
        in the OpticStudio <pop> folder. Only used when beam_type is set to 'File', 'DLL' or 'Multimode'. Defaults to
        ''.
    beam_parameters : dict | None
        The beam parameters. Either a dictionary containing the beam parameters settings or None. Only the parameters
        that are defined in the dictionary will be set, the other parameters will remain their default value. Note that
        the accepted beam parameters vary per beam type. The accepted parameters for a specific beam type and their
        default values can be obtained through zospy.analysis.physicaloptics.pop_create_beam_parameter_dict(), which can
        then be adjusted and parsed to beam_parameters. When set to None, the default beam parameters will be used.
        Defaults to None.
    x_sampling : constants.Analysis.SampleSizes | str | int
        The number of points used to sample the beam in the x direction. Either specified using
        constants.Analysis.SampleSizes, a string (e.g. '32x32') or integer (e.g. 32). Note that the string or integer
        should match one of the sample sizes in constants.Analysis.SampleSizes (32 or '32x32' for 'S_32x32' etc.).
        Defaults to 32.
    y_sampling : constants.Analysis.SampleSizes | str | int
        The number of points used to sample the beam in the y direction. Either specified using
        constants.Analysis.SampleSizes, a string (e.g. '32x32') or integer (e.g. 32). Note that the string or integer
        should match one of the sample sizes in constants.Analysis.SampleSizes (32 or '32x32' for 'S_32x32' etc.).
        Defaults to 32.
    x_width : float
        The original width of the region represented by the array in OpticStudio lens units. Defaults to 1.0.
    y_width : float
        The original width of the region represented by the array in OpticStudio lens units. Defaults to 1.0.
    use_total_power : bool
        Use total power. Note that either use_total_power or use_peak_irradiance has to be True. Defaults to False.
    total_power : float
        The total power. Only used when use_total_power is True. Defaults to 1.0.
    use_peak_irradiance : bool
        Use peak irradiance. Note that either use_total_power or use_peak_irradiance has to be True. Defaults to True.
    peak_irradiance : float
        The peak irradiance. Only used when use_peak_irradiance is True. Defaults to 1.0.
    show_as : constants.Analysis.PhysicalOptics.POPShowAsTypes | str
        Defines the default graph in OpticStudio. Defaults to 'FalseColor'.
    data_type : constants.Analysis.PhysicalOptics.POPDataTypes | str
        Defines the type of output data. Defaults to 'Irradiance'.
    project : constants.Analysis.PhysicalOptics.POPProjectionTypes | str
        Defines the perspective from which the beam is viewed. Defaults to 'AlongBeam'.
    contour_format : str
        The contour format string. An empty string means no formatting. Only used when show_as is set to 'Contour'.
        Defaults to ''.
    plot_scale : float
        The plot scale. Defaults to 0.
    scale_type : constants.Analysis.PhysicalOptics.POPScaleTypes | str
        The scale type for the plot. Defaults to 'Linear'.
    zoom_in : constants.Analysis.PhysicalOptics.POPZoomTypes | str
        The zoom factor for the plot. Defaults to 'NoZoom'.
    row_or_column : int | str
        An integer defining the row or column number used when show_as is either 'CrossX' or 'CrossY'. Also accepts
        'Center'. Defaults to 'Center'.
    zero_phase_level : float
        The lower limit on the relative irradiance. Data points below this threshold have a phase value of 0. Defaults
        to 0.001.
    save_output_beam : bool
        Defines whether the output beam should be saved. Defaults to False.
    output_beam_file : str
        The filename to which the output beam should be saved. Should NOT include the .ZBF extension. The file will be
        saved in the OpticStudio <pop> folder.
    save_beam_at_all_surfaces : bool
        Defines whether the beam should be saved at all surfaces. Only used when save_output_beam is set to True.
        Defaults to False.
    compute_fiber_coupling_integral : bool
        Defines whether the fibre coupling is computed. Defaults to False.
    ignore_fiber_polarization :
        Defines whether polarization is considered. Defaults to True.
    fiber_type : constants.Analysis.PhysicalOptics.POPFiberTypes | str
        The fiber type. Defaults to 'GaussianWaist'.
    fiber_position : constants.Analysis.PhysicalOptics.POPFiberPositions | str
        The reference for the center of the receiving fiber. Defaults t 'ChiefRay'.
    tilt_about_x :
        The tilt of the fiber around the x-axis in degrees with respect to the beam. Defaults to 0.0.
    tilt_about_y :
        The tilt of the fiber around the y-axis in degrees with respect to the beam. Defaults to 0.0.
    fiber_type_file :
        The file containing the fiber description. Only used is fiber_type is set to 'File' or 'DLL'.
    fiber_parameters : dict | None
        The fiber parameters. Either a dictionary containing the fiber parameters settings or None. Only the parameters
        that are defined in the dictionary will be set, the other parameters will remain their default value. Note that
        the accepted beam parameters vary per fiber type. The accepted parameters for a specific fiber type and their
        default values can be obtained through zospy.analysis.physicaloptics.pop_create_fiber_parameter_dict(), which
        can then be adjusted and parsed to fiber_parameters. When set to None, the default fiber parameters will be
        used. Defaults to None.
    auto_calculate_beam_sampling : bool
        Defines whether the beam sampling should be automatically calculated. Will adjust x_width and y_width. This
        calculation is done just before the analysis is run. Only used when beam_type is not one of ('File', 'DLL',
        'Multimode'). Defaults to False.
    """

    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number")
    field: int = Field(default=1, ge=1, description="Field number")
    start_surface: Literal["Ent. Pupil"] | Annotated[int, Field(ge=1)] = Field(default=1, description="Start surface")
    end_surface: Literal["Image"] | Annotated[int, Field(ge=1)] = Field(default="Image", description="End surface")
    surface_to_beam: float = Field(default=0, description="Surface to beam")
    use_polarization: bool = Field(default=False, description="Use polarization")
    separate_xy: bool = Field(default=False, description="Separate propagation in X Y direction")
    use_disk_storage: bool = Field(default=False, description="Use disk storage")

    # Beam definition
    beam_type: ZOSAPIConstant("Analysis.PhysicalOptics.POPBeamTypes") = Field(
        default="GaussianWaist", description="Beam type"
    )
    beam_file: str = Field(default="", description="Beam file")
    beam_parameters: dict | None = Field(default=None, description="Beam parameters dictionary")
    x_sampling: str | Annotated[int, Field(ge=1)] = Field(default=32, description="Beam sampling in X direction")
    y_sampling: str | Annotated[int, Field(ge=1)] = Field(default=32, description="Beam sampling in Y direction")
    x_width: float = Field(default=1, description="Beam width in X direction")
    y_width: float = Field(default=1, description="Beam width in Y direction")
    use_total_power: bool = Field(default=False, description="Use total power")
    total_power: float = Field(default=1, description="Total power")
    use_peak_irradiance: bool = Field(default=True, description="Use peak irradiance")
    peak_irradiance: float = Field(default=1, description="Peak irradiance")

    # Display
    show_as: ZOSAPIConstant("Analysis.PhysicalOptics.POPShowAsTypes") = Field(
        default="FalseColor", description="Show as"
    )
    data_type: ZOSAPIConstant("Analysis.PhysicalOptics.POPDataTypes") = Field(
        default="Irradiance", description="Data type"
    )
    project: ZOSAPIConstant("Analysis.PhysicalOptics.POPProjectionTypes") = Field(
        default="AlongBeam", description="Projection types"
    )
    contour_format: str = Field(default="", description="Contour format")
    plot_scale: float = Field(default=0, description="Plot scale")
    scale_type: ZOSAPIConstant("Analysis.PhysicalOptics.POPScaleTypes") = Field(
        default="Linear", description="Scale type"
    )
    zoom_in: ZOSAPIConstant("Analysis.PhysicalOptics.POPZoomTypes") = Field(default="NoZoom", description="Zoom in")
    row_or_column: int | Literal["Center"] = Field(default="Center", description="Row or column")
    zero_phase_level: float = Field(default=0.001, description="Zero phase level")
    save_output_beam: bool = Field(default=False, description="Save output beam")
    output_beam_file: str = Field(default="", description="Output beam file")
    save_beam_at_all_surfaces: bool = Field(default=False, description="Save beam at all surfaces")

    # Fiber definition
    compute_fiber_coupling_integral: bool = Field(default=False, description="Compute fiber coupling integral")
    ignore_fiber_polarization: bool = Field(default=True, description="Ignore fibre polarization")
    fiber_type: ZOSAPIConstant("Analysis.PhysicalOptics.POPFiberTypes") = Field(
        default="GaussianWaist", description="Fiber type"
    )
    fiber_position: ZOSAPIConstant("Analysis.PhysicalOptics.POPFiberPositions") = Field(
        default="ChiefRay", description="Fiber position"
    )
    tilt_about_x: float = Field(default=0, description="Tilt about X")
    tilt_about_y: float = Field(default=0, description="Tilt about Y")
    fiber_type_file: str = Field(default="", description="Fiber type file")
    fiber_parameters: dict | None = Field(default=None, description="Fiber parameters dictionary")

    auto_calculate_beam_sampling: bool = Field(default=False, description="Automatically calculate beam sampling")

    @model_validator(mode="after")
    def _validate_total_power_peak_irradiance(self):
        if not (self.use_total_power ^ self.use_peak_irradiance):
            raise ValueError(
                "Either use_total_power or use_peak_irradiance should be True, they cannot both be True or False."
            )

        return self

    @model_validator(mode="after")
    def _validate_beam_file(self):
        if str(self.beam_type) not in ("File", "DLL", "Multimode") and self.beam_file != "":
            raise ValueError(f"Beam type {self.beam_type!s} does not allow specification of beam_file.")

        return self

    @model_validator(mode="after")
    def _validate_output_beam_file(self):
        if self.save_output_beam and self.output_beam_file == "":
            raise ValueError("output_beam_file should be specified when save_output_beam is True.")
        if not self.save_output_beam:
            if self.output_beam_file != "":
                warn("output_beam_file is ignored when save_output_beam is False.")
            if self.save_beam_at_all_surfaces:
                warn("save_beam_at_all_surfaces is ignored when save_output_beam is False.")

        return self

    @model_validator(mode="after")
    def _validate_fiber_type_file(self):
        if str(self.fiber_type) not in ("File", "DLL") and self.fiber_type_file != "":
            raise ValueError(f"Fiber type {self.fiber_type!s} does not allow specification of fiber_type_file.")

        return self


class PhysicalOpticsPropagation(
    BaseAnalysisWrapper[Union[DataFrame, None], PhysicalOpticsPropagationSettings],
    analysis_type="PhysicalOpticsPropagation",
    mode="Sequential",
):
    """Physical Optics Propagation analysis."""

    def __init__(
        self,
        *,
        wavelength: WavelengthNumber = "All",
        field: int = 1,
        start_surface: Literal["Ent. Pupil"] | int = 1,
        end_surface: Literal["Image"] | int = "Image",
        surface_to_beam: float = 0,
        use_polarization: bool = False,
        separate_xy: bool = False,
        use_disk_storage: bool = False,
        beam_type: constants.Analysis.PhysicalOptics.POPBeamTypes | str = "GaussianWaist",
        beam_file: str = "",
        beam_parameters: dict | None = None,
        x_sampling: constants.Analysis.SampleSizes | str | int = 32,
        y_sampling: constants.Analysis.SampleSizes | str | int = 32,
        x_width: float = 1,
        y_width: float = 1,
        use_total_power: bool = False,
        total_power: float = 1,
        use_peak_irradiance: bool = True,
        peak_irradiance: float = 1,
        show_as: constants.Analysis.PhysicalOptics.POPShowAsTypes | str = "FalseColor",
        data_type: constants.Analysis.PhysicalOptics.POPDataTypes | str = "Irradiance",
        project: constants.Analysis.PhysicalOptics.POPProjectionTypes | str = "AlongBeam",
        contour_format: str = "",
        plot_scale: float = 0,
        scale_type: constants.Analysis.PhysicalOptics.POPScaleTypes | str = "Linear",
        zoom_in: constants.Analysis.PhysicalOptics.POPZoomTypes | str = "NoZoom",
        row_or_column: int | str = "Center",
        zero_phase_level: float = 0.001,
        save_output_beam: bool = False,
        output_beam_file: str = "",
        save_beam_at_all_surfaces: bool = False,
        compute_fiber_coupling_integral: bool = False,
        ignore_fiber_polarization: bool = True,
        fiber_type: constants.Analysis.PhysicalOptics.POPFiberTypes | str = "GaussianWaist",
        fiber_position: constants.Analysis.PhysicalOptics.POPFiberPositions | str = "ChiefRay",
        tilt_about_x: float = 0,
        tilt_about_y: float = 0,
        fiber_type_file: str = "",
        fiber_parameters: dict | None = None,
        auto_calculate_beam_sampling: bool = False,
        settings: PhysicalOpticsPropagationSettings | None = None,
    ):
        """Create a Physical Optics Propagation analysis.

        See Also
        --------
        PhysicalOpticsPropagationSettings : Settings for the Physical Optics Propagation analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> DataFrame:
        """Run the Physical Optics Propagation analysis."""
        if self.settings.start_surface == "Ent. Pupil":
            self.analysis.Settings.StartSurface.SetSurfaceNumber(0)
        elif isinstance(self.settings.start_surface, int):
            self.analysis.Settings.StartSurface.SetSurfaceNumber(self.settings.start_surface)
        else:
            raise ValueError(
                f'start_surface value should be "Ent. Pupil" or an integer, got {self.settings.start_surface}'
            )

        if self.settings.end_surface == "Image":
            self.analysis.Settings.EndSurface.UseImageSurface()
        elif isinstance(self.settings.end_surface, int):
            self.analysis.Settings.EndSurface.SetSurfaceNumber(self.settings.end_surface)
        else:
            raise ValueError(f'end_surface value should be "Image" or an integer, got {self.settings.end_surface}')

        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.SurfaceToBeam = self.settings.surface_to_beam
        self.analysis.Settings.UsePolarization = self.settings.use_polarization
        self.analysis.Settings.SeparateXY = self.settings.separate_xy
        self.analysis.Settings.UseDiskStorage = self.settings.use_disk_storage

        # Beam settings
        self.analysis.Settings.BeamType = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPBeamTypes, self.settings.beam_type
        )

        x_sampling = (
            f"{self.settings.x_sampling}x{self.settings.y_sampling}"
            if isinstance(self.settings.x_sampling, int)
            else self.settings.x_sampling
        )
        y_sampling = (
            f"{self.settings.x_sampling}x{self.settings.y_sampling}"
            if isinstance(self.settings.y_sampling, int)
            else self.settings.y_sampling
        )

        self.analysis.Settings.XSampling = constants.process_constant(
            constants.Analysis.SampleSizes, standardize_sampling(x_sampling)
        )
        self.analysis.Settings.YSampling = constants.process_constant(
            constants.Analysis.SampleSizes, standardize_sampling(y_sampling)
        )
        self.analysis.Settings.XWidth = self.settings.x_width
        self.analysis.Settings.YWidth = self.settings.y_width

        if self.settings.use_total_power:
            self.analysis.Settings.UseTotalPower = True
            self.analysis.Settings.TotalPower = self.settings.total_power
        else:
            self.analysis.Settings.UsePeakIrradiance = True
            self.analysis.Settings.PeakIrradiance = self.settings.peak_irradiance

        if self.settings.beam_file != "":
            self.analysis.Settings.BeamTypeFilename = self.settings.beam_file

        self._set_pop_parameters("beam", self.settings.beam_parameters)

        # Display settings
        self.analysis.Settings.ShowAs = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPShowAsTypes, self.settings.show_as
        )
        self.analysis.Settings.DataType = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPDataTypes, self.settings.data_type
        )
        self.analysis.Settings.Project = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPProjectionTypes, self.settings.project
        )

        if self.settings.show_as == "Contour":
            self.analysis.Settings.ContourFormat = self.settings.contour_format
        elif self.settings.show_as in ("CrossX", "CrossY"):
            self.analysis.Settings.RowOrColumn = (
                0 if self.settings.row_or_column == "Center" else self.settings.row_or_column
            )

        self.analysis.Settings.PlotScale = self.settings.plot_scale
        self.analysis.Settings.LogScale = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPScaleTypes, self.settings.scale_type
        )
        self.analysis.Settings.ZoomIn = constants.process_constant(
            constants.Analysis.PhysicalOptics.POPZoomTypes, self.settings.zoom_in
        )
        self.analysis.Settings.ZeroPhaseLevel = self.settings.zero_phase_level

        self.analysis.Settings.SaveOutputBeam = self.settings.save_output_beam
        if self.settings.save_output_beam:
            self.analysis.Settings.OutputBeamFile = self.settings.output_beam_file
            self.analysis.Settings.SaveBeamAtAllSurfaces = self.settings.save_beam_at_all_surfaces

        # Fiber settings
        self.analysis.Settings.ComputeFiberCouplingIntegral = self.settings.compute_fiber_coupling_integral

        if self.settings.compute_fiber_coupling_integral:
            self.analysis.Settings.IgnoreFiberPolarization = self.settings.ignore_fiber_polarization
            self.analysis.Settings.FiberType = constants.process_constant(
                constants.Analysis.PhysicalOptics.POPFiberTypes, self.settings.fiber_type
            )
            self.analysis.Settings.FiberPosition = constants.process_constant(
                constants.Analysis.PhysicalOptics.POPFiberPositions, self.settings.fiber_position
            )
            self.analysis.Settings.TiltAboutX = self.settings.tilt_about_x
            self.analysis.Settings.TiltAboutY = self.settings.tilt_about_y

            if self.settings.fiber_type_file != "":
                self.analysis.Settings.FiberTypeFilename = self.settings.fiber_type_file

            self._set_pop_parameters("fiber", self.settings.fiber_parameters)

        # Auto calculate beam sampling
        if self.settings.auto_calculate_beam_sampling and str(self.settings.beam_type) not in (
            "File",
            "DLL",
            "Multimode",
        ):
            self.analysis.Settings.AutoCalculateBeamSampling()

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        return self.get_data_grid(cell_origin="bottom_left")

    def _set_pop_parameters(self, which: Literal["beam", "fiber"], parameters: dict) -> None:
        """Set beam or fiber parameters using the provided dictionary.

        Only parameters present in the dictionary are set. After setting, the function checks if all provided parameters
        were used and raises an error if any are left unused.

        Parameters
        ----------
        analysis : Analysis
            The analysis object containing settings for either beam or fiber parameters.
        which : Literal['beam', 'fiber']
            Specifies whether to set 'beam' or 'fiber' parameters. Determines which set of methods to use within the
            `analysis` object.
        parameters : dict | None
            A dictionary containing parameter names and their desired values. See also pop_create_beam_parameter_dict() and
            pop_create_fiber_parameter_dict(). Parameters not accepted by the current beam or fiber type will raise an
            error.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If `which` is not 'beam' or 'fiber', or if any parameters in `parameters` do not match those expected for the
            specified `which` type in `analysis`.
        """
        if parameters is not None:
            parameters = parameters.copy()  # Create copy to preserve the original dictionary

            if which == "beam":
                number_of_parameters = self.analysis.Settings.NumberOfParameters
                get_parameter_name = self.analysis.Settings.GetParameterName
                set_parameter_value = self.analysis.Settings.SetParameterValue
            elif which == "fiber":
                number_of_parameters = self.analysis.Settings.NumberOfFiberParameters
                get_parameter_name = self.analysis.Settings.GetFiberParameterName
                set_parameter_value = self.analysis.Settings.SetFiberParameterValue
            else:
                raise ValueError("Invalid parameter type. Choose 'fiber' or 'beam'.")

            for i in range(number_of_parameters):
                param_name = get_parameter_name(i)
                param_value = parameters.pop(param_name, None)
                if param_value is not None:  # Only set parameters that are present in the dictionary
                    set_parameter_value(i, param_value)

            if parameters:
                raise ValueError(
                    f"The following {which} parameters are specified but not accepted: {parameters.keys()}. "
                    f"The accepted {which} parameters for this {which}_type are: "
                    f"{','.join([get_parameter_name(i) for i in range(number_of_parameters)])}"
                )


def create_beam_parameter_dict(
    oss: OpticStudioSystem, beam_type: constants.Analysis.PhysicalOptics.POPBeamTypes | str = "GaussianWaist"
) -> dict:
    """Create a dictionary containing the parameters for a certain Physical Optics beam type and their default values.

    The dictionary can be adjusted and supplied to the beam_parameters argument of physical_optics_propagation()

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    beam_type : constants.Analysis.PhysicalOptics.POPBeamTypes | str
        The beam type. Defaults to 'GaussianWaist'.

    Returns
    -------
    dict
        A dictionary containing the parameters for a certain Physical Optics beam type and their default values.
    """
    analysis_type = constants.Analysis.AnalysisIDM.PhysicalOpticsPropagation

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    analysis.Settings.BeamType = constants.process_constant(constants.Analysis.PhysicalOptics.POPBeamTypes, beam_type)

    beam_params = {}
    for i in range(analysis.Settings.NumberOfParameters):
        beam_params[analysis.Settings.GetParameterName(i)] = analysis.Settings.GetParameterValue(i)

    analysis.Close()

    return beam_params


def create_fiber_parameter_dict(
    oss: OpticStudioSystem, fiber_type: constants.Analysis.PhysicalOptics.POPFiberTypes | str = "GaussianWaist"
) -> dict:
    """Create a dictionary containing the parameters for a certain Physical Optics fiber type and their default values.

     The dictionary can be adjusted and supplied to the fiber_parameters argument of physical_optics_propagation()

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    fiber_type : constants.Analysis.PhysicalOptics.POPFiberTypes | str
         The fiber type. Defaults to 'GaussianWaist'.

    Returns
    -------
    dict
        A dictionary containing the parameters for a certain Physical Optics fiber type and their default values.
    """
    analysis_type = constants.Analysis.AnalysisIDM.PhysicalOpticsPropagation

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    analysis.Settings.FiberType = constants.process_constant(
        constants.Analysis.PhysicalOptics.POPFiberTypes, fiber_type
    )

    fiber_params = {}
    for i in range(analysis.Settings.NumberOfFiberParameters):
        fiber_params[analysis.Settings.GetFiberParameterName(i)] = analysis.Settings.GetFiberParameterValue(i)

    analysis.Close()

    return fiber_params



"""OpticStudio Solvers.

`zospy.solvers` provides easy access to various solvers offered by OpticStudio.
Solvers have the general structure `solver_function(cell, *args)`, with `cell` the EditorCell of the solver,
followed by the parameters for the solver.

Examples
--------
Make the radius of a surface variable:

>>> surface = oss.InsertNewSurfaceAt(1)
>>> zp.solvers.variable(surface.RadiusCell)

Change the refractive index of a medium:

>>> surface = oss.InsertNewSurfaceAt(1)
>>> zp.solvers.material_model(surface.MaterialCell, refractive_index=1.5)
"""

from __future__ import annotations

from zospy.api import _ZOSAPI, constants

__all__ = ("element_power", "fixed", "material_model", "pickup_chief_ray", "position", "surface_pickup", "variable")

# ruff: noqa: SLF001


def _get_surface_index(surface: _ZOSAPI.Editors.IEditorRow | int) -> int:
    if isinstance(surface, int):
        return surface

    if surface.GetType().BaseType.FullName == "ZemaxUI.ZOSAPI.Editors.ZOSAPI_EditorRowBase":
        return surface.RowIndex

    raise ValueError(f"from_surface should be an int or a Surface, got {surface}")


def element_power(radius_cell: _ZOSAPI.Editors.IEditorCell, power: float) -> _ZOSAPI.Editors.ISolveElementPower:
    """Solver for element power.

    Adjusts the value of `radius_cell` to create an element with the specified `power`. This solver should be set on
    the last surface of the element.

    Parameters
    ----------
    radius_cell : ZOSAPI.Editors.IEditorCell
        Radius cell of the last element surface
    power : float
        Element power in diopters

    Returns
    -------
    solve_data : ZOSAPI.Editors.ISolveElementPower
        The solve data for the element power
    """
    solve_data = radius_cell.CreateSolveType(constants.Editors.SolveType.ElementPower)._S_ElementPower

    # The ZOS-API accepts powers in diopters instead of diopters / 1000
    solve_data.Power = power

    radius_cell.SetSolveData(solve_data)

    return solve_data


def fixed(cell: _ZOSAPI.Editors.IEditorCell) -> _ZOSAPI.Editors.ISolveFixed:
    """Set the cell solve type to Fixed.

    Parameters
    ----------
    cell : ZOSAPI.Editors.IEditorCell

    Returns
    -------
    The solve data object for `cell`.
    """
    solve_data = cell.CreateSolveType(constants.Editors.SolveType.Fixed)._S_Fixed
    cell.SetSolveData(solve_data)

    return solve_data


def material_model(
    material_cell: _ZOSAPI.Editors.IEditorCell,
    refractive_index: float = 1,
    abbe_number: float = 0,
    partial_dispersion: float = 0,
) -> _ZOSAPI.Editors.ISolveMaterialModel:
    """Solver for material model.

    Sets the material type of a specific LDE surface to model. This solver only works on material cells.

    Parameters
    ----------
    material_cell : ZOSAPI.Editors.IEditorCell
        MaterialCell of the LDE surface row or NCE object row that requires the modification
    refractive_index : float
        Refractive index of the material
    abbe_number : float
        Abbe number of the material
    partial_dispersion : float
        Partial dispersion term at d-light

    Returns
    -------
    The SolveData object for the material model.
    """
    solve_data = material_cell.CreateSolveType(constants.Editors.SolveType.MaterialModel)._S_MaterialModel
    solve_data.IndexNd = refractive_index
    solve_data.AbbeVd = abbe_number
    solve_data.dPgF = partial_dispersion

    material_cell.SetSolveData(solve_data)

    return solve_data


def pickup_chief_ray(
    cell: _ZOSAPI.Editors.IEditorCell, field: int = 1, wavelength: int = 0
) -> _ZOSAPI.Editors.ISolvePickupChiefRay:
    """Set the cell solve type to Chief Ray Pickup.

    Adjusts the value of `cell` so the chief ray passes through the center of the surface.

    Parameters
    ----------
    cell : ZOSAPI.Editors.IEditorCell
    field : int, optional
        Field index to take the chief ray from, defaults to 1.
    wavelength : int, optional
        Wavelength index of the chief ray, defaults to 0.

    Returns
    -------
    ZOSAPI.Editors.ISolvePickupChiefRay
        The solve data for the chief ray pickup
    """
    solve_data = cell.CreateSolveType(constants.Editors.SolveType.PickupChiefRay)._S_PickupChiefRay

    solve_data.Field = field
    solve_data.Wavelength = wavelength

    cell.SetSolveData(solve_data)

    return solve_data


def position(
    thickness_cell: _ZOSAPI.Editors.IEditorCell,
    from_surface: _ZOSAPI.Editors.LDE.ILDERow | _ZOSAPI.Editors.NCE.INCERow,
    length: float,
) -> _ZOSAPI.Editors.ISolvePosition:
    """Set the cell solve type to Position.

    Solves the position of `surface` relative to `from_surface`. This solver works only on thickness cells.

    Parameters
    ----------
    thickness_cell : ZOSAPI.Editors.IEditorCell
        Surface for which the position should be set
    from_surface : ZOSAPI.Editors.IEditorRow | int
        Index of the surface or surface from which the position is measured
    length : float
        Distance between `from_surface` and `surface`.

    Returns
    -------
        The SolveData object for the position.
    """
    solve_data = thickness_cell.CreateSolveType(constants.Editors.SolveType.Position)._S_Position

    if isinstance(from_surface, int):
        solve_data.FromSurface = from_surface
    elif type(from_surface).__name__ in ("ILDERow", "INCERow"):
        solve_data.FromSurface = from_surface.RowIndex
    else:
        raise ValueError(f"from_surface should be an int or a Surface, got {from_surface}")

    solve_data.Length = length

    thickness_cell.SetSolveData(solve_data)

    return solve_data


def surface_pickup(
    cell: _ZOSAPI.Editors.IEditorCell,
    from_surface: _ZOSAPI.Editors.IEditorRow | int,
    from_column: constants.Editors.LDE.SurfaceColumn | str = None,
    scale: float = 1,
    offset: float = 0,
) -> _ZOSAPI.Editors.ISolveSurfacePickup:
    """Set the cell solve type to Surface Pickup.

    This solver picks up the value from another surface and optionally applies a scale and offset to it.

    Parameters
    ----------
    cell : ZOSAPI.Editors.IEditorCell
        Cell for which the value should be set
    from_surface : ZOSAPI.Editors.IEditorRow
        Index of the surface or surface from which the value is picked up
    from_column : zospy.api.constants.Editors.LDE.SurfaceColumn | str, optional
        Column from which the value is picked up
    scale : float
        Factor by which the picked up value is scaled. This value is only set if the cell supports a scale
        factor. Defaults to 1.
    offset : float
        Offset which is added to the picked up value. This value is only set if the cell supports an offset.
        Defaults to 0.

    Returns
    -------
        The SolveData object for the surface pickup.
    """
    solve_data = cell.CreateSolveType(constants.Editors.SolveType.SurfacePickup)._S_SurfacePickup

    solve_data.Surface = _get_surface_index(from_surface)

    if solve_data.SupportsScale:
        solve_data.ScaleFactor = scale

    if solve_data.SupportsOffset:
        solve_data.Offset = offset

    if from_column is not None:
        solve_data.Column = constants.process_constant(constants.Editors.LDE.SurfaceColumn, from_column)

    cell.SetSolveData(solve_data)

    return solve_data


def variable(cell: _ZOSAPI.Editors.IEditorCell) -> _ZOSAPI.Editors.ISolveVariable:
    """Set the cell solve type to Variable.

    Parameters
    ----------
    cell : ZOSAPI.Editors.IEditorCell

    Returns
    -------
    The solve data object for `cell`.
    """
    solve_data = cell.CreateSolveType(constants.Editors.SolveType.Variable)._S_Variable
    cell.SetSolveData(solve_data)

    return solve_data



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import (
    BestFitSphereOptions,
    RemoveOptions,
    SampleSizes_Pow2Plus1,
    SampleSizes_Pow2Plus1_X,
    ShowAs,
    SurfaceCurvatureCrossData,
    SurfaceCurvatureData,
    SurfacePhaseData,
    SurfacePhaseSlopeCrossData,
    SurfacePhaseSlopeData,
    SurfaceSagData,
    SurfaceSlopeCrossData,
    SurfaceSlopeData,
)
from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = (
    "IAS_SurfaceCurvature",
    "IAS_SurfaceCurvatureCross",
    "IAS_SurfacePhase",
    "IAS_SurfacePhaseCross",
    "IAS_SurfacePhaseSlope",
    "IAS_SurfacePhaseSlopeCross",
    "IAS_SurfaceSag",
    "IAS_SurfaceSagCross",
    "IAS_SurfaceSlope",
    "IAS_SurfaceSlopeCross",
)

class IAS_SurfaceCurvature(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfaceCurvatureData: ...
    @Data.setter
    def Data(self, value: SurfaceCurvatureData) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_SurfaceCurvatureCross(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Data(self) -> SurfaceCurvatureCrossData: ...
    @Data.setter
    def Data(self, value: SurfaceCurvatureCrossData) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1) -> None: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_SurfacePhase(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfacePhaseData: ...
    @Data.setter
    def Data(self, value: SurfacePhaseData) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemovePiston(self) -> bool: ...
    @RemovePiston.setter
    def RemovePiston(self, value: bool) -> None: ...
    @property
    def RemoveTiltX(self) -> bool: ...
    @RemoveTiltX.setter
    def RemoveTiltX(self, value: bool) -> None: ...
    @property
    def RemoveTiltY(self) -> bool: ...
    @RemoveTiltY.setter
    def RemoveTiltY(self, value: bool) -> None: ...
    @property
    def RemovePower(self) -> bool: ...
    @RemovePower.setter
    def RemovePower(self, value: bool) -> None: ...

class IAS_SurfacePhaseCross(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Data(self) -> SurfacePhaseData: ...
    @Data.setter
    def Data(self, value: SurfacePhaseData) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1) -> None: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def RemovePiston(self) -> bool: ...
    @RemovePiston.setter
    def RemovePiston(self, value: bool) -> None: ...
    @property
    def RemoveTiltX(self) -> bool: ...
    @RemoveTiltX.setter
    def RemoveTiltX(self, value: bool) -> None: ...
    @property
    def RemoveTiltY(self) -> bool: ...
    @RemoveTiltY.setter
    def RemoveTiltY(self, value: bool) -> None: ...
    @property
    def RemovePower(self) -> bool: ...
    @RemovePower.setter
    def RemovePower(self, value: bool) -> None: ...

class IAS_SurfacePhaseSlope(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfacePhaseSlopeData: ...
    @Data.setter
    def Data(self, value: SurfacePhaseSlopeData) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemovePiston(self) -> bool: ...
    @RemovePiston.setter
    def RemovePiston(self, value: bool) -> None: ...
    @property
    def RemoveTiltX(self) -> bool: ...
    @RemoveTiltX.setter
    def RemoveTiltX(self, value: bool) -> None: ...
    @property
    def RemoveTiltY(self) -> bool: ...
    @RemoveTiltY.setter
    def RemoveTiltY(self, value: bool) -> None: ...
    @property
    def RemovePower(self) -> bool: ...
    @RemovePower.setter
    def RemovePower(self, value: bool) -> None: ...

class IAS_SurfacePhaseSlopeCross(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Data(self) -> SurfacePhaseSlopeCrossData: ...
    @Data.setter
    def Data(self, value: SurfacePhaseSlopeCrossData) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1) -> None: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def RemovePiston(self) -> bool: ...
    @RemovePiston.setter
    def RemovePiston(self, value: bool) -> None: ...
    @property
    def RemoveTiltX(self) -> bool: ...
    @RemoveTiltX.setter
    def RemoveTiltX(self, value: bool) -> None: ...
    @property
    def RemoveTiltY(self) -> bool: ...
    @RemoveTiltY.setter
    def RemoveTiltY(self, value: bool) -> None: ...
    @property
    def RemovePower(self) -> bool: ...
    @RemovePower.setter
    def RemovePower(self, value: bool) -> None: ...

class IAS_SurfaceSag(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfaceSagData: ...
    @Data.setter
    def Data(self, value: SurfaceSagData) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_SurfaceSagCross(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Data(self) -> SurfaceSagData: ...
    @Data.setter
    def Data(self, value: SurfaceSagData) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1) -> None: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_SurfaceSlope(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfaceSlopeData: ...
    @Data.setter
    def Data(self, value: SurfaceSlopeData) -> None: ...
    @property
    def ShowAs(self) -> ShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: ShowAs) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...

class IAS_SurfaceSlopeCross(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def Data(self) -> SurfaceSlopeCrossData: ...
    @Data.setter
    def Data(self, value: SurfaceSlopeCrossData) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1) -> None: ...
    @property
    def Angle(self) -> float: ...
    @Angle.setter
    def Angle(self, value: float) -> None: ...
    @property
    def RemoveOption(self) -> RemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: RemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def ConsiderOffAxisAperture(self) -> bool: ...
    @ConsiderOffAxisAperture.setter
    def ConsiderOffAxisAperture(self, value: bool) -> None: ...
    @property
    def ReverseDirection(self) -> bool: ...
    @ReverseDirection.setter
    def ReverseDirection(self, value: bool) -> None: ...



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("ColorRaysBy", "Patterns", "Reference", "ShowScales")

class ColorRaysBy:
    Fields: ColorRaysBy = None
    Waves: ColorRaysBy = None
    Config: ColorRaysBy = None
    Wavelength: ColorRaysBy = None

class Patterns:
    Square: Patterns = None
    Hexapolar: Patterns = None
    Dithered: Patterns = None

class Reference:
    ChiefRay: Reference = None
    Centroid: Reference = None
    Middle: Reference = None
    Vertex: Reference = None

class ShowScales:
    ScaleBar: ShowScales = None
    Box: ShowScales = None
    Cross: ShowScales = None
    Circle: ShowScales = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("HuygensMtfTypes", "MtfDataTypes", "MtfTypes", "ShowAsTypes", "SurfaceMtfTypes")

class HuygensMtfTypes:
    Modulation: HuygensMtfTypes = None

class MtfDataTypes:
    Average: MtfDataTypes = None
    Tangential: MtfDataTypes = None
    Sagittal: MtfDataTypes = None

class MtfTypes:
    Modulation: MtfTypes = None
    Real: MtfTypes = None
    Imaginary: MtfTypes = None
    Phase: MtfTypes = None
    SquareWave: MtfTypes = None

class ShowAsTypes:
    GreyScale: ShowAsTypes = None
    GreyScaleInverted: ShowAsTypes = None
    FalseColor: ShowAsTypes = None
    FalseColorInverted: ShowAsTypes = None

class SurfaceMtfTypes:
    MTF_Incoherent: SurfaceMtfTypes = None
    Real_Incoherent: SurfaceMtfTypes = None
    Imaginary_Incoherent: SurfaceMtfTypes = None
    MTF_Coherent: SurfaceMtfTypes = None
    Real_Coherent: SurfaceMtfTypes = None
    Imaginary_Coherent: SurfaceMtfTypes = None



"""Zemax OpticStudio analyses from the Extended Scene Analysis category."""

from __future__ import annotations

from typing import Literal

import pandas as pd

from zospy import utils
from zospy.analyses.old.base import AnalysisResult, OnComplete, new_analysis
from zospy.api import constants
from zospy.zpcore import OpticStudioSystem


def geometric_image_analysis(
    oss: OpticStudioSystem,
    wavelength: Literal["All"] | int = "All",
    field: int = 1,
    surface: Literal["Image"] | int = "Image",
    field_size: float = 0,
    image_size: float = 50,
    file: str = "LETTERF.IMA",
    rotation: float = 0,
    rays_x_1000: int = 10,
    show_as: constants.Analysis.GiaShowAsTypes | str = "Surface",
    source: constants.Analysis.Settings.SourceGia | str = "Uniform",
    number_of_pixels: int = 100,
    row_column_number: Literal["Center"] | int = "Center",
    na: float = 0,
    total_watts: float = 1,
    parity: constants.Analysis.Settings.Parity | str = "Even",
    reference: constants.Analysis.Settings.ReferenceGia | str = "ChiefRay",
    use_symbols: bool = False,
    use_polarization: bool = False,
    remove_vignetting_factors: bool = True,
    scatter_rays: bool = False,
    delete_vignetted: bool = False,
    use_pixel_interpolation: bool = False,
    save_as_bim_file: str = "",
    oncomplete: OnComplete | str = OnComplete.Close,
) -> AnalysisResult:
    """Geometric image analysis.

    See the OpticStudio documentation for a more in depth description of most parameters.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    wavelength : str | int
        The wavelength to use in the analysis. Either 'All' or an integer specifying the wavelength number. Defaults to
        'All'.
    field : int
        The field to use in the analysis. Should be an integer specifying the field number. Defaults to 1.
    surface : str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    field_size : float
        The width of the image in field coordinates. Defaults to 0.
    image_size : float
        When show_as is set to 'SpotDiagram', this defines the size of the scale bar superimposed on the image. Else, it
         defines the size of the detector used to capture the rays. Defaults to 50.
    file : str
        The image file used for the analysis. Must be an .IMA of .BIM file and reside in the OpticStudio images folder.
        Defaults to 'LETTERF.IMA'.
    rotation : float
        The rotation angle in degrees. Defaults to 0.
    rays_x_1000 : int
        The approximate amount of rays that are traced. See OpticStudio manual for further clarification. Defaults to
        10.
    show_as : constants.Analysis.GiaShowAsTypes | str
        Defines how the result is displayed in OpticStudio. Defaults to 'Surface'. Note that 'SpotDiagram' is not
        implmented.
    source : constants.Analysis.Settings.SourceGia | str
        Specifies how the source is defined. Defaults to 'Uniform'.
    number_of_pixels : int
        The number of pixels across the width of the image. Not used when show_as is set to 'SpotDiagram'. Defaults to
        100.
    row_column_number : int | str
        An integer defining the row or column number used when show_as is either 'CrossX' or 'CrossY'. Also accepts
        'Center'. Defaults to 'Center'.
    na : float
        The numerical aperture cut-off. Ignored if set to 0. Defaults to 0.
    total_watts : float
        The total power in watts. Defaults to 1.
    parity : constants.Analysis.Settings.Parity | str
        Defines how the object would appear in object space. Defaults to 'Even'.
    reference : constants.Analysis.Settings.ReferenceGia | str
        Defines the reference coordinate for the center of the plot. Defaults to 'ChiefRay'.
    use_symbols : bool
        Defines whether the plot uses symbols. Only used when show-as is set to 'SpotDiagram'. Defaults to False.
    use_polarization : bool
        Defines whether polarization is considered. Defaults to False.
    remove_vignetting_factors : bool
        Defines whether vignetting factors are automatically removed. Defaults to True.
    scatter_rays : bool
        Defines whether rays are scattered by surfaces with scattering properties. Defaults to False.
    delete_vignetted : False
        Defines whether rays vignetted by any surface are drawn. Defaults to False.
    use_pixel_interpolation : bool
        Defines whether pixel interpolation is used. This setting has no effect when show-as is set to 'SpotDiagram'.
        Defaults to False.
    save_as_bim_file : str
        Defines the filename used to save the output image as BIM file. This file will be stored in the OpticStudio
        images folder. The filename should include the .BIM extension. If set to '', the image will not be saved. Not
        used when show_as is set to 'SpotDiagram.  Defaults to ''.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.

    Returns
    -------
    AnalysisResult
        A Geometric Image Analysis result.

    Raises
    ------
    NotImplementedError
        When show_as is set to SpotDiagram.
    """
    if str(show_as) == "SpotDiagram":
        raise NotImplementedError("Geometric Image Analysis with show_as set to SpotDiagram is not implemented.")

    analysis_type = constants.Analysis.AnalysisIDM.GeometricImageAnalysis

    # Create analysis
    analysis = new_analysis(oss, analysis_type)

    # Adjust settings
    analysis.set_wavelength(wavelength)
    analysis.set_field(field)
    analysis.set_surface(surface)
    analysis.Settings.FieldSize = field_size
    analysis.Settings.ImageSize = image_size
    analysis.Settings.File = file
    analysis.Settings.Rotation = rotation
    analysis.Settings.RaysX1000 = rays_x_1000
    analysis.Settings.ShowAs = constants.process_constant(constants.Analysis.GiaShowAsTypes, show_as)
    analysis.Settings.Source = constants.process_constant(constants.Analysis.Settings.SourceGia, source)

    if str(show_as) in ("CrossX", "CrossY"):
        if row_column_number == "Center":
            analysis.Settings.UseColumnRowCenter()
        else:
            analysis.Settings.RowColumnNumber = row_column_number
    if str(show_as) != "SpotDiagram":
        analysis.Settings.NumberOfPixels = number_of_pixels
    if str(show_as) == "SpotDiagram":
        analysis.Settings.UseSymbols = use_symbols

    analysis.Settings.NA = na
    analysis.Settings.TotalWatts = total_watts
    analysis.Settings.Parity = constants.process_constant(constants.Analysis.Settings.Parity, parity)
    analysis.Settings.Reference = constants.process_constant(constants.Analysis.Settings.ReferenceGia, reference)

    analysis.Settings.UsePolarization = use_polarization
    analysis.Settings.RemoveVignettingFactors = remove_vignetting_factors
    analysis.Settings.ScatterRays = scatter_rays
    analysis.Settings.DeleteVignetted = delete_vignetted
    analysis.Settings.UsePixelInterpolation = use_pixel_interpolation
    analysis.Settings.SaveAsBIMFile = save_as_bim_file

    # Calculate
    analysis.ApplyAndWaitForCompletion()

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)
    settings.loc["Field"] = analysis.get_field()
    settings.loc["Wavelength"] = analysis.get_wavelength()
    settings.loc["Surface"] = analysis.Settings.Surface.GetSurfaceNumber()
    settings.loc["FieldSize"] = analysis.Settings.FieldSize
    settings.loc["ImageSize"] = analysis.Settings.ImageSize
    settings.loc["File"] = analysis.Settings.File
    settings.loc["Rotation"] = analysis.Settings.Rotation
    settings.loc["RaysX1000"] = analysis.Settings.RaysX1000
    settings.loc["ShowAs"] = str(analysis.Settings.ShowAs)
    settings.loc["Source"] = str(analysis.Settings.Source)
    if str(analysis.Settings.ShowAs) != "SpotDiagram":
        settings.loc["NumberOfPixels"] = analysis.Settings.NumberOfPixels
    if str(analysis.Settings.ShowAs) in ("CrossX", "CrossY"):
        if analysis.Settings.RowColumnNumber == 0:
            settings.loc["Row"] = "Center"
        else:
            settings.loc["Row"] = analysis.Settings.RowColumnNumber
    settings.loc["NA"] = analysis.Settings.NA
    settings.loc["TotalWatts"] = analysis.Settings.TotalWatts
    settings.loc["Parity"] = str(analysis.Settings.Parity)
    settings.loc["Reference"] = str(analysis.Settings.Reference)
    if str(analysis.Settings.ShowAs) == "SpotDiagram":
        settings.loc["UseSymbols"] = analysis.Settings.UseSymbols
    settings.loc["UsePolarization"] = analysis.Settings.UsePolarization
    settings.loc["RemoveVignettingFactors"] = analysis.Settings.RemoveVignettingFactors
    settings.loc["ScatterRays"] = analysis.Settings.ScatterRays
    settings.loc["DeleteVignetted"] = analysis.Settings.DeleteVignetted
    settings.loc["UsePixelInterpolation"] = analysis.Settings.UsePixelInterpolation
    settings.loc["SaveAsBIMFile"] = analysis.Settings.SaveAsBIMFile

    # Get data and unpack
    data = []
    for ii in range(analysis.Results.NumberOfDataGrids):
        data.append(utils.zputils.unpack_datagrid(analysis.Results.DataGrids[ii],
                                                  # pass "center" and None for consistency with older ZOSPy versions
                                                  cell_origin="center", label_rounding=None))

    if len(data) == 0:
        data = pd.DataFrame()
    elif len(data) == 1:
        data = data[0]
    else:
        data = pd.concat(data, axis=1)

    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)



"""OpticStudio analyses from the Reports category."""

from zospy.analyses.reports.cardinal_points import CardinalPoints, CardinalPointsSettings
from zospy.analyses.reports.surface_data import SurfaceData, SurfaceDataSettings
from zospy.analyses.reports.system_data import SystemData

__all__ = ("SurfaceData", "SurfaceDataSettings", "SystemData", "CardinalPoints", "CardinalPointsSettings")



from types import SimpleNamespace

import pytest
from semver import Version

from zospy.analyses.old.systemviewers import (
    cross_section,
    nsc_3d_layout,
    nsc_shaded_model,
    shaded_model,
    viewer_3d,
)

# ruff: noqa: SLF001

pytestmark = pytest.mark.old_analyses


class TestViewers:
    def test_can_run_cross_section(self, simple_system):
        result = cross_section(simple_system, oncomplete="Sustain")
        analysistype = str(result.Analysis._analysis.AnalysisType)
        result.Analysis._analysis.Close()

        assert analysistype == "Draw2D"

    def test_can_run_viewer_3d(self, simple_system):
        result = viewer_3d(simple_system, oncomplete="Sustain")
        analysistype = str(result.Analysis._analysis.AnalysisType)
        result.Analysis._analysis.Close()

        assert analysistype == "Draw3D"

    def test_can_run_shaded_model(self, simple_system):
        result = shaded_model(simple_system, oncomplete="Sustain")
        analysistype = str(result.Analysis._analysis.AnalysisType)
        result.Analysis._analysis.Close()

        assert analysistype == "ShadedModel"

    def test_can_run_nsc_3d_layout(self, nsc_simple_system):
        result = nsc_3d_layout(nsc_simple_system, oncomplete="Sustain")
        analysistype = str(result.Analysis._analysis.AnalysisType)
        result.Analysis._analysis.Close()

        assert analysistype == "NSC3DLayout"

    def test_can_run_nsc_shaded_model(self, nsc_simple_system):
        result = nsc_shaded_model(nsc_simple_system, oncomplete="Sustain")
        analysistype = str(result.Analysis._analysis.AnalysisType)
        result.Analysis._analysis.Close()

        assert analysistype == "NSCShadedModel"

    @pytest.mark.parametrize(
        "analysis,parameters",
        [
            (cross_section, {"start_surface": 2, "number_of_rays": 10}),
            (viewer_3d, {"end_surface": 2, "hide_x_bars": True}),
        ],
    )
    def test_old_opticstudio_version_raises_warning(self, simple_system, analysis, parameters, monkeypatch):
        monkeypatch.setattr(simple_system, "ZOS", SimpleNamespace(version=Version(20, 1, 0)))

        with pytest.warns(UserWarning, match=", ".join(parameters.keys())):
            analysis(simple_system, **parameters, oncomplete="Close")



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import BestFitSphereOptions, SampleSizes_Pow2Plus1_X, SurfaceSagData
from zospy.api._ZOSAPI.Analysis.Settings import IAS_

__all__ = ("IAS_NSCSurfaceSag", "NSCSagRemoveOptions", "NSCSagShowAs")

class IAS_NSCSurfaceSag(IAS_):
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def ObjectNumber(self) -> int: ...
    @ObjectNumber.setter
    def ObjectNumber(self, value: int) -> None: ...
    @property
    def FaceNumber(self) -> int: ...
    @FaceNumber.setter
    def FaceNumber(self, value: int) -> None: ...
    @property
    def Sampling(self) -> SampleSizes_Pow2Plus1_X: ...
    @Sampling.setter
    def Sampling(self, value: SampleSizes_Pow2Plus1_X) -> None: ...
    @property
    def Data(self) -> SurfaceSagData: ...
    @Data.setter
    def Data(self, value: SurfaceSagData) -> None: ...
    @property
    def ShowAs(self) -> NSCSagShowAs: ...
    @ShowAs.setter
    def ShowAs(self, value: NSCSagShowAs) -> None: ...
    @property
    def Offset(self) -> float: ...
    @Offset.setter
    def Offset(self, value: float) -> None: ...
    @property
    def Width(self) -> float: ...
    @Width.setter
    def Width(self, value: float) -> None: ...
    @property
    def Rotation(self) -> float: ...
    @Rotation.setter
    def Rotation(self, value: float) -> None: ...
    @property
    def Decenter_X(self) -> float: ...
    @Decenter_X.setter
    def Decenter_X(self, value: float) -> None: ...
    @property
    def Decenter_Y(self) -> float: ...
    @Decenter_Y.setter
    def Decenter_Y(self, value: float) -> None: ...
    @property
    def ContourFormat(self) -> str: ...
    @ContourFormat.setter
    def ContourFormat(self, value: str) -> None: ...
    @property
    def RemoveOption(self) -> NSCSagRemoveOptions: ...
    @RemoveOption.setter
    def RemoveOption(self, value: NSCSagRemoveOptions) -> None: ...
    @property
    def BestFitSphereOption(self) -> BestFitSphereOptions: ...
    @BestFitSphereOption.setter
    def BestFitSphereOption(self, value: BestFitSphereOptions) -> None: ...
    @property
    def BFSReverseDirection(self) -> bool: ...
    @BFSReverseDirection.setter
    def BFSReverseDirection(self, value: bool) -> None: ...
    @property
    def KeepObjectTilts(self) -> bool: ...
    @KeepObjectTilts.setter
    def KeepObjectTilts(self, value: bool) -> None: ...
    @property
    def ProbeRayOffset(self) -> float: ...
    @ProbeRayOffset.setter
    def ProbeRayOffset(self, value: float) -> None: ...
    @property
    def KeepZRD(self) -> bool: ...
    @KeepZRD.setter
    def KeepZRD(self, value: bool) -> None: ...
    @property
    def ZRDFilename(self) -> str: ...
    @ZRDFilename.setter
    def ZRDFilename(self, value: str) -> None: ...

class NSCSagRemoveOptions:
    # None = 0
    BaseROC = 1
    BestFitSphere = 2
    AverageSag = 3
    MinimumSag = 4

class NSCSagShowAs:
    Surface = 0
    Contour = 1
    GreyScale = 2
    InverseGreyScale = 3
    FalseColor = 4
    InverseFalseColor = 5
    X_CrossSection = 6
    Y_CrossSection = 7



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations
from . import Analysis, Common, Editors, Preferences, SystemData, Tools, Wizards

__all__ = (
    "Analysis",
    "Common",
    "Editors",
    "Preferences",
    "SystemData",
    "Tools",
    "Wizards",
    "LensUpdateMode",
    "LicenseStatusType",
    "SessionModes",
    "STARUpdateMode",
    "SystemType",
    "UpdateStatus",
    "ZOSAPI_Mode",
)

class LensUpdateMode:
    None_: LensUpdateMode = None
    EditorsOnly: LensUpdateMode = None
    AllWindows: LensUpdateMode = None

class LicenseStatusType:
    Unknown: LicenseStatusType = None
    KeyNotWorking: LicenseStatusType = None
    NewLicenseNeeded: LicenseStatusType = None
    StandardEdition: LicenseStatusType = None
    ProfessionalEdition: LicenseStatusType = None
    PremiumEdition: LicenseStatusType = None
    TooManyInstances: LicenseStatusType = None
    NotAuthorized: LicenseStatusType = None
    KeyNotFound: LicenseStatusType = None
    KeyExpired: LicenseStatusType = None
    Timeout: LicenseStatusType = None
    InstanceConflict: LicenseStatusType = None
    OpticsViewer: LicenseStatusType = None
    OpticStudioHPCEdition: LicenseStatusType = None
    EnterpriseEdition: LicenseStatusType = None
    StudentEdition: LicenseStatusType = None

class SessionModes:
    FromPreferences: SessionModes = None
    SessionOn: SessionModes = None
    SessionOff: SessionModes = None

class STARUpdateMode:
    Normal: STARUpdateMode = None
    Suspended: STARUpdateMode = None

class SystemType:
    Sequential: SystemType = None
    NonSequential: SystemType = None

class UpdateStatus:
    NotChecked: UpdateStatus = None
    UpToDate: UpdateStatus = None
    AvailableEligible: UpdateStatus = None
    AvailableIneligible: UpdateStatus = None
    CheckFailed: UpdateStatus = None
    NotSupported: UpdateStatus = None

class ZOSAPI_Mode:
    Server: ZOSAPI_Mode = None
    Operand: ZOSAPI_Mode = None
    Plugin: ZOSAPI_Mode = None
    UserAnalysis: ZOSAPI_Mode = None
    UserAnalysisSettings: ZOSAPI_Mode = None



"""OpticStudio system viewers."""

from zospy.analyses.systemviewers.cross_section import CrossSection, CrossSectionSettings
from zospy.analyses.systemviewers.nsc_3d_layout import NSC3DLayout, NSC3DLayoutSettings
from zospy.analyses.systemviewers.nsc_shaded_model import NSCShadedModel, NSCShadedModelSettings
from zospy.analyses.systemviewers.shaded_model import ShadedModel, ShadedModelSettings
from zospy.analyses.systemviewers.viewer_3d import Viewer3D, Viewer3DSettings

__all__ = (
    "CrossSection",
    "CrossSectionSettings",
    "Viewer3D",
    "Viewer3DSettings",
    "ShadedModel",
    "ShadedModelSettings",
    "NSC3DLayout",
    "NSC3DLayoutSettings",
    "NSCShadedModel",
    "NSCShadedModelSettings",
)



"""OpticStudio analyses from the Polarization category."""

from zospy.analyses.polarization.pupil_map import PolarizationPupilMap, PolarizationPupilMapSettings
from zospy.analyses.polarization.transmission import (
    PolarizationTransmission,
    PolarizationTransmissionSettings,
)

__all__ = (
    "PolarizationTransmission",
    "PolarizationTransmissionSettings",
    "PolarizationPupilMap",
    "PolarizationPupilMapSettings",
)



"""Surface Data analysis."""

from __future__ import annotations

from pathlib import Path  # noqa: TCH003
from typing import Annotated, Any

from pydantic import AliasChoices, BeforeValidator, Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.transformers import SimpleField, ZospyTransformer

__all__ = ("SurfaceData", "SurfaceDataSettings")


class SurfaceDataTransformer(ZospyTransformer):
    """Transformer for the output of the Surface Data analysis."""

    def refractive_index_table(self, args):
        """Convert the refractive index table to a list of dictionaries."""
        header, rows = args[0]
        table_records = [dict(zip(header, row)) for row in rows]

        return SimpleField("Refractive Indices", table_records)

    def surface_number(self, n):
        """Extract the surface number from the header."""
        return SimpleField("Surface Number", n)

    def surface_powers(self, args):
        """Convert the surface powers to a dictionary."""
        return SimpleField("Surface Powers", dict(args))


@analysis_result
class EdgeThickness:
    y: float = Field(alias="Y Edge Thick")
    x: float = Field(alias="X Edge Thick")


@analysis_result
class ModelGlass:
    nd: float = Field(alias="nd")
    abbe: float = Field(alias="Abbe")
    dpgf: float = Field(alias="dPgF")


@analysis_result
class RefractiveIndex:
    number: int = Field(alias="#")
    wavelength: float = Field(alias="Wavelength")
    index: float = Field(alias="Index")


@analysis_result
class MaterialData:
    indices: list[RefractiveIndex] = Field(alias="Refractive Indices")
    best_fit_glass: str | None = Field(alias="Best Fit Glass", default=None)
    glass: str | ModelGlass | None = Field(alias=AliasChoices("Model glass", "Glass"), default=None)


def _deserialize_tuple(value: Any) -> tuple[float, ...] | Any:
    """Deserialize a tuple from a string.

    Used for deserialization of JSON-serialized analysis results.
    """
    if isinstance(value, str):
        return tuple(map(float, value.split(",")))

    return value


TupleKey = Annotated[tuple[float, float], BeforeValidator(_deserialize_tuple)]


@analysis_result
class SurfacePower:
    surf: dict[int, float] = Field(alias="Surf")
    power: dict[TupleKey, float] | None = Field(alias="Power", default=None)
    efl: dict[TupleKey, float] | None = Field(alias="EFL", default=None)
    f_number: dict[TupleKey, float] | None = Field(alias="F/#", default=None)


@analysis_result
class SurfacePowers:
    as_situated: SurfacePower = Field(alias="as situated")
    in_air: SurfacePower = Field(alias="in air")


@analysis_result
class SurfaceDataResult:
    """Data for the Surface Data analysis."""

    comment: str | None = Field(alias="Comment", default=None)
    date: str = Field(alias="Date")
    file: Path = Field(alias="File")
    diameter: float = Field(alias="Diameter")
    edge_thickness: EdgeThickness = Field(alias="Edge Thickness")
    material: MaterialData = Field(alias="Index of Refraction")
    lens_units: str = Field(alias="Lens units")
    surface_powers: SurfacePowers = Field(alias="Surface Powers")
    thickness: float = Field(alias="Thickness")
    shape_factor: float | None = Field(alias="Shape Factor", default=None)


@analysis_settings
class SurfaceDataSettings:
    """Settings for the Surface Data analysis.

    Attributes
    ----------
    surface : int
        The surface number that is to be analyzed
    """

    surface: int = Field(default=1, ge=0, description="Surface number to analyze.")


class SurfaceData(
    BaseAnalysisWrapper[SurfaceDataResult, SurfaceDataSettings],
    analysis_type="SurfaceDataSettings",
    needs_text_output_file=True,
    needs_config_file=True,
):
    """Surface Data analysis."""

    def __init__(self, *, surface: int = 1):
        """Create a new Surface Data analysis.

        See Also
        --------
        SurfaceDataSettings : Settings for the Surface Data analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> SurfaceDataResult:
        """Run the Surface Data analysis."""
        settings = self.analysis.GetSettings()
        settings.SaveTo(str(self.config_file))

        settings_bytestring = self.config_file.read_bytes()
        settings_bytearray = bytearray(settings_bytestring)
        settings_bytearray[20] = self.settings.surface  # 20 maps to the selected surface

        self.config_file.write_bytes(settings_bytearray)

        settings.LoadFrom(str(self.config_file))

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Read text file and parse to object
        return self.parse_output(
            "surface_data",
            transformer=SurfaceDataTransformer,
            result_type=SurfaceDataResult,
        )



import numpy as np
import pytest

from zospy.analyses.old.wavefront import wavefront_map, zernike_standard_coefficients

pytestmark = pytest.mark.old_analyses


class TestWavefrontMap:
    def test_can_run_wavefront_map(self, simple_system):
        result = wavefront_map(simple_system)

        assert result.Data is not None

    def test_to_json(self, simple_system):
        result = wavefront_map(simple_system)

        assert result.from_json(result.to_json())

    @pytest.mark.parametrize(
        "sampling,use_exit_pupil", [("64x64", True), ("64x64", False), ("128x128", True), ("128x128", False)]
    )
    def test_wavefront_map_returns_correct_result(self, simple_system, sampling, use_exit_pupil, expected_data):
        result = wavefront_map(simple_system, sampling=sampling, use_exit_pupil=use_exit_pupil)

        assert np.allclose(result.Data.to_numpy(dtype=float), expected_data.Data.to_numpy(dtype=float), equal_nan=True)

    @pytest.mark.parametrize(
        "sampling,use_exit_pupil", [("64x64", True), ("64x64", False), ("128x128", True), ("128x128", False)]
    )
    def test_wavefront_map_matches_reference_data(self, simple_system, sampling, use_exit_pupil, reference_data):
        result = wavefront_map(simple_system, sampling=sampling, use_exit_pupil=use_exit_pupil)

        assert np.allclose(result.Data.to_numpy(dtype=float), reference_data.Data.to_numpy(dtype=float), equal_nan=True)


class TestZernikeStandardCoefficients:
    def test_can_run_zernike_standard_coefficients(self, simple_system):
        result = zernike_standard_coefficients(simple_system)

        assert result.Data is not None

    def test_to_json(self, simple_system):
        result = zernike_standard_coefficients(simple_system)

        assert result.from_json(result.to_json())

    @pytest.mark.parametrize("sampling,maximum_term", [("64x64", 37), ("128x128", 64)])
    def test_zernike_standard_coefficients_returns_correct_result(
        self, simple_system, sampling, maximum_term, expected_data
    ):
        result = zernike_standard_coefficients(simple_system, sampling=sampling, maximum_term=maximum_term)

        assert np.allclose(
            result.Data.Coefficients.Value.astype(float), expected_data.Data.Coefficients.Value.astype(float), atol=1e-8
        )
        assert np.isclose(
            result.Data.GeneralData.loc["StrehlRatio(Est)", "Value"],
            expected_data.Data.GeneralData.loc["StrehlRatio(Est)", "Value"],
        )
        assert all(result.Data.Coefficients.Function == expected_data.Data.Coefficients.Function)

    @pytest.mark.parametrize("sampling,maximum_term", [("64x64", 37), ("128x128", 64)])
    def test_zernike_standard_coefficients_matches_reference_data(
        self, simple_system, sampling, maximum_term, reference_data
    ):
        result = zernike_standard_coefficients(simple_system, sampling=sampling, maximum_term=maximum_term)

        assert np.allclose(
            result.Data.Coefficients.Value.astype(float),
            reference_data.Data.Coefficients.Value.astype(float),
            atol=1e-8,
        )
        assert np.isclose(
            result.Data.GeneralData.loc["StrehlRatio(Est)", "Value"],
            reference_data.Data.GeneralData.loc["StrehlRatio(Est)", "Value"],
        )
        assert all(result.Data.Coefficients.Function == reference_data.Data.Coefficients.Function)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from typing import overload

from zospy.api._ZOSAPI.Analysis import ErrorType
from zospy.api._ZOSAPI.Tools import ISystemTool

__all__ = (
    "BatchRayTraceDataEntry",
    "BatchRayTraceDataHolder",
    "IBatchRayTrace",
    "ILightningTrace",
    "INSCRayTrace",
    "IRayTraceDirectPolData",
    "IRayTraceDirectUnpolData",
    "IRayTraceNormPolData",
    "IRayTraceNormUnpolData",
    "IRayTraceNSCData",
    "IRayTraceNSCSourceData",
    "IZRDReader",
    "IZRDReaderResults",
    "LTEdgeSasmpling",
    "LTRaySampling",
    "NSCTraceOptions",
    "OPDMode",
    "RayPathDataEntry",
    "RayPathDataHolder",
    "RayStatus",
    "RaysType",
    "ZRDFormatType",
)

class BatchRayTraceDataEntry: ...

class BatchRayTraceDataHolder:
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, data: list[BatchRayTraceDataEntry]): ...
    @overload
    def __init__(self, info: SerializationInfo, context: StreamingContext): ...
    def GetObjectData(self, info: SerializationInfo, context: StreamingContext) -> None: ...

class IBatchRayTrace(ISystemTool):
    def CreateDirectPol(
        self,
        MaxRays: int,
        rayType: RaysType,
        Ex: float,
        Ey: float,
        phax: float,
        phay: float,
        startSurface: int,
        toSurface: int,
    ) -> IRayTraceDirectPolData: ...
    def CreateDirectUnpol(
        self, MaxRays: int, rayType: RaysType, startSurface: int, toSurface: int
    ) -> IRayTraceDirectUnpolData: ...
    def CreateNormPol(
        self, MaxRays: int, rayType: RaysType, Ex: float, Ey: float, phaX: float, phaY: float, toSurface: int
    ) -> IRayTraceNormPolData: ...
    def CreateNormUnpol(self, MaxRays: int, rayType: RaysType, toSurface: int) -> IRayTraceNormUnpolData: ...
    def CreateNSC(self, MaxRays: int, maxSegments: int, coherenceLength: float) -> IRayTraceNSCData: ...
    def CreateNSCSourceData(self, maxSegments: int, coherenceLength: float) -> IRayTraceNSCSourceData: ...
    def GetDirectFieldCoordinates(
        self, waveNumber: int, rayType: RaysType, Hx: float, Hy: float, Px: float, Py: float
    ) -> tuple[bool, float, float, float, float, float, float]: ...
    def GetPhase(
        self, L: float, M: float, N: float, jx: float, jy: float, xPhaseDeg: float, yPhaseDeg: float, intensity: float
    ) -> tuple[float, float, float, float, float, float]: ...
    def SingleRayDirectPol(
        self,
        rayType: RaysType,
        Ex: float,
        Ey: float,
        phaX: float,
        phaY: float,
        startSurface: int,
        toSurface: int,
        waveNumber: int,
        X: float,
        Y: float,
        Z: float,
        L: float,
        M: float,
        N: float,
    ) -> tuple[bool, int, int, float, float, float, float, float, float, float]: ...
    def SingleRayDirectPolFull(
        self,
        rayType: RaysType,
        Ex: float,
        Ey: float,
        phaX: float,
        phaY: float,
        startSurface: int,
        toSurface: int,
        waveNumber: int,
        X: float,
        Y: float,
        Z: float,
        L: float,
        M: float,
        N: float,
    ) -> tuple[
        bool, int, int, float, float, float, float, float, float, float, float, float, float, float, float, float
    ]: ...
    def SingleRayDirectUnpol(
        self,
        rayType: RaysType,
        startSurface: int,
        toSurface: int,
        waveNumber: int,
        X: float,
        Y: float,
        Z: float,
        L: float,
        M: float,
        N: float,
    ) -> tuple[bool, int, int, float, float, float, float, float, float, float, float, float, float]: ...
    def SingleRayNormPol(
        self,
        rayType: RaysType,
        Ex: float,
        Ey: float,
        phaX: float,
        phaY: float,
        toSurf: int,
        waveNumber: int,
        Hx: float,
        Hy: float,
        Px: float,
        Py: float,
        exr: float,
        exi: float,
        eyr: float,
        eyi: float,
        ezr: float,
        ezi: float,
    ) -> tuple[bool, int, float, float, float, float, float, float, float]: ...
    def SingleRayNormPolFull(
        self,
        rayType: RaysType,
        Ex: float,
        Ey: float,
        phaX: float,
        phaY: float,
        toSurf: int,
        waveNumber: int,
        Hx: float,
        Hy: float,
        Px: float,
        Py: float,
        exr: float,
        exi: float,
        eyr: float,
        eyi: float,
        ezr: float,
        ezi: float,
    ) -> tuple[
        bool, int, float, float, float, float, float, float, float, float, float, float, float, float, float
    ]: ...
    def SingleRayNormUnpol(
        self, rayType: RaysType, toSurf: int, waveNumber: int, Hx: float, Hy: float, Px: float, Py: float, calcOPD: bool
    ) -> tuple[bool, int, int, float, float, float, float, float, float, float, float, float, float, float]: ...

class ILightningTrace(ISystemTool):
    @property
    def RaySampling(self) -> LTRaySampling: ...
    @RaySampling.setter
    def RaySampling(self, value: LTRaySampling) -> None: ...
    @property
    def EdgeSampling(self) -> LTEdgeSasmpling: ...
    @EdgeSampling.setter
    def EdgeSampling(self, value: LTEdgeSasmpling) -> None: ...

class INSCRayTrace(ISystemTool):
    @property
    def AutoUpdate(self) -> bool: ...
    @AutoUpdate.setter
    def AutoUpdate(self, value: bool) -> None: ...
    @property
    def NumberOfCores(self) -> int: ...
    @NumberOfCores.setter
    def NumberOfCores(self, value: int) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def IgnoreErrors(self) -> bool: ...
    @IgnoreErrors.setter
    def IgnoreErrors(self, value: bool) -> None: ...
    @property
    def SplitNSCRays(self) -> bool: ...
    @SplitNSCRays.setter
    def SplitNSCRays(self, value: bool) -> None: ...
    @property
    def ScatterNSCRays(self) -> bool: ...
    @ScatterNSCRays.setter
    def ScatterNSCRays(self, value: bool) -> None: ...
    @property
    def SaveRays(self) -> bool: ...
    @SaveRays.setter
    def SaveRays(self, value: bool) -> None: ...
    @property
    def SaveRaysFile(self) -> str: ...
    @SaveRaysFile.setter
    def SaveRaysFile(self, value: str) -> None: ...
    @property
    def SavePaths(self) -> bool: ...
    @SavePaths.setter
    def SavePaths(self, value: bool) -> None: ...
    @property
    def SavePathsFile(self) -> str: ...
    @SavePathsFile.setter
    def SavePathsFile(self, value: str) -> None: ...
    @property
    def ZRDFormat(self) -> ZRDFormatType: ...
    @ZRDFormat.setter
    def ZRDFormat(self, value: ZRDFormatType) -> None: ...
    @property
    def Filter(self) -> str: ...
    @Filter.setter
    def Filter(self, value: str) -> None: ...
    @property
    def DeadRayErrors(self) -> float: ...
    @property
    def DeadRayThreshold(self) -> float: ...
    @property
    def IsHPCEnabled(self) -> bool: ...
    @property
    def HPCEstimatedTimeS(self) -> float: ...
    @property
    def HPCRemainingTimeS(self) -> float: ...
    @property
    def HPCTimeToStartS(self) -> float: ...
    @property
    def HPCQueuePosition(self) -> int: ...
    @property
    def HPCState(self) -> HPCRunState: ...
    @property
    def HPCHasTimeEstimate(self) -> bool: ...
    @property
    def RayMultiplier(self) -> float: ...
    @RayMultiplier.setter
    def RayMultiplier(self, value: float) -> None: ...
    @property
    def HasHPCUnsupportedDetectors(self) -> bool: ...
    def ClearDetectorObject(self, ObjectNumber: int) -> ErrorType: ...
    def ClearDetectors(self, DetectorNumber: int) -> ErrorType: ...
    def EstimateHPCTime(self) -> bool: ...
    def GetTotalRayEnergy(self) -> float: ...
    def ResetRandomSeed(self) -> None: ...
    def SetPartialRTMode(self, seed: int, group: int) -> None: ...
    def SetRandomSeed(self, seed: int) -> None: ...

class IRayTraceDirectPolData:
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def MaxRays(self) -> int: ...
    @property
    def HasResultData(self) -> bool: ...
    def AddRay(self, waveNumber: int, X: float, Y: float, Z: float, L: float, M: float, N: float) -> bool: ...
    def ClearData(self) -> None: ...
    def ReadNextResult(self) -> tuple[bool, int, int, int, float, float, float, float, float, float, float]: ...
    def ReadNextResultFull(
        self,
    ) -> tuple[
        bool, int, int, int, float, float, float, float, float, float, float, float, float, float, float, float, float
    ]: ...
    def StartReadingResults(self) -> bool: ...

class IRayTraceDirectUnpolData:
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def MaxRays(self) -> int: ...
    @property
    def HasResultData(self) -> bool: ...
    def AddRay(self, waveNumber: int, X: float, Y: float, Z: float, L: float, M: float, N: float) -> bool: ...
    def ClearData(self) -> None: ...
    def ReadNextResult(
        self,
    ) -> tuple[bool, int, int, int, float, float, float, float, float, float, float, float, float, float]: ...
    def StartReadingResults(self) -> bool: ...

class IRayTraceNormPolData:
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def MaxRays(self) -> int: ...
    @property
    def HasResultData(self) -> bool: ...
    def AddRay(
        self,
        waveNumber: int,
        Hx: float,
        Hy: float,
        Px: float,
        Py: float,
        exr: float,
        exi: float,
        eyr: float,
        eyi: float,
        ezr: float,
        ezi: float,
    ) -> bool: ...
    def ClearData(self) -> None: ...
    def ReadNextResult(self) -> tuple[bool, int, int, float, float, float, float, float, float, float]: ...
    def ReadNextResultFull(
        self,
    ) -> tuple[
        bool, int, int, float, float, float, float, float, float, float, float, float, float, float, float, float
    ]: ...
    def StartReadingResults(self) -> bool: ...

class IRayTraceNormUnpolData:
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def MaxRays(self) -> int: ...
    @property
    def HasResultData(self) -> bool: ...
    def AddRay(self, waveNumber: int, Hx: float, Hy: float, Px: float, Py: float, calcOPD: OPDMode) -> bool: ...
    def ClearData(self) -> None: ...
    def ReadNextResult(
        self,
    ) -> tuple[bool, int, int, int, float, float, float, float, float, float, float, float, float, float, float]: ...
    def StartReadingResults(self) -> bool: ...

class IRayTraceNSCData:
    @property
    def NumberOfRays(self) -> int: ...
    @property
    def MaxRays(self) -> int: ...
    @property
    def HasResultData(self) -> bool: ...
    def AddRay(
        self,
        waveNumber: int,
        surf: int,
        mode: NSCTraceOptions,
        X: float,
        Y: float,
        Z: float,
        L: float,
        M: float,
        N: float,
        InsideOf: int,
        exr: float,
        exi: float,
        eyr: float,
        eyi: float,
        ezr: float,
        ezi: float,
    ) -> bool: ...
    def ClearData(self) -> None: ...
    def ReadNextResult(self) -> tuple[bool, int, int, int, int]: ...
    def ReadNextSegment(
        self,
    ) -> tuple[
        bool,
        int,
        int,
        int,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]: ...
    def StartReadingResults(self) -> bool: ...

class IRayTraceNSCSourceData:
    @property
    def UseSingleSource(self) -> bool: ...
    @UseSingleSource.setter
    def UseSingleSource(self, value: bool) -> None: ...
    @property
    def SurfaceNumber(self) -> int: ...
    @SurfaceNumber.setter
    def SurfaceNumber(self, value: int) -> None: ...
    @property
    def ObjectNumber(self) -> int: ...
    @ObjectNumber.setter
    def ObjectNumber(self, value: int) -> None: ...
    @property
    def MaxRays(self) -> int: ...
    @MaxRays.setter
    def MaxRays(self, value: int) -> None: ...
    @property
    def TraceOptions(self) -> NSCTraceOptions: ...
    @TraceOptions.setter
    def TraceOptions(self, value: NSCTraceOptions) -> None: ...
    @property
    def Wavelength(self) -> int: ...
    @Wavelength.setter
    def Wavelength(self, value: int) -> None: ...
    @property
    def HasResultData(self) -> bool: ...
    def ReadNextResult(self) -> tuple[bool, int, int, int, int]: ...
    def ReadNextSegment(
        self,
    ) -> tuple[
        bool,
        int,
        int,
        int,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]: ...
    def StartReadingResults(self) -> bool: ...
    def UseAnyWavelength(self) -> None: ...
    def UsePrimaryWavelength(self) -> None: ...

class IZRDReader(ISystemTool):
    @property
    def ZRDFile(self) -> None: ...
    @ZRDFile.setter
    def ZRDFile(self, value: str) -> None: ...
    @property
    def Filter(self) -> str: ...
    @Filter.setter
    def Filter(self, value: str) -> None: ...
    def GetResults(self) -> IZRDReaderResults: ...

class IZRDReaderResults:
    @property
    def IsValid(self) -> bool: ...
    @property
    def RayNumber(self) -> int: ...
    @property
    def WaveIndex(self) -> int: ...
    @property
    def WlUM(self) -> float: ...
    @property
    def NumSegments(self) -> int: ...
    @property
    def SegTerminated(self) -> bool: ...
    @property
    def SegReflected(self) -> bool: ...
    @property
    def SegTransmitted(self) -> bool: ...
    @property
    def SegScattered(self) -> bool: ...
    @property
    def SegDiffracted(self) -> bool: ...
    @property
    def SegBulkScattered(self) -> bool: ...
    @property
    def SegmentTIR(self) -> bool: ...
    @property
    def ParentGhost(self) -> bool: ...
    @property
    def ParentDiffracted(self) -> bool: ...
    @property
    def ParentScattered(self) -> bool: ...
    @property
    def RayError(self) -> bool: ...
    @property
    def SegmentLevel(self) -> int: ...
    @property
    def SegmentParent(self) -> int: ...
    @property
    def HitObject(self) -> int: ...
    @property
    def HitFace(self) -> int: ...
    @property
    def InsideOf(self) -> int: ...
    @property
    def X(self) -> float: ...
    @property
    def Y(self) -> float: ...
    @property
    def Z(self) -> float: ...
    @property
    def L(self) -> float: ...
    @property
    def M(self) -> float: ...
    @property
    def N(self) -> float: ...
    @property
    def EXR(self) -> float: ...
    @property
    def EXI(self) -> float: ...
    @property
    def EYR(self) -> float: ...
    @property
    def EYI(self) -> float: ...
    @property
    def EZR(self) -> float: ...
    @property
    def EZI(self) -> float: ...
    @property
    def Intensity(self) -> float: ...
    @property
    def PathLength(self) -> float: ...
    @property
    def OpticalPathW(self) -> float: ...
    @property
    def OpticalPathLU(self) -> float: ...
    @property
    def XYBin(self) -> float: ...
    @property
    def LMBin(self) -> float: ...
    @property
    def XNorm(self) -> float: ...
    @property
    def YNorm(self) -> float: ...
    @property
    def ZNorm(self) -> float: ...
    @property
    def Index(self) -> float: ...
    @property
    def StartingPhase(self) -> float: ...
    @property
    def PhaseOf(self) -> float: ...
    @property
    def PhaseAt(self) -> float: ...
    @property
    def HasOrderX(self) -> bool: ...
    @property
    def HasOrderY(self) -> bool: ...
    @property
    def OrderX(self) -> int: ...
    @property
    def OrderY(self) -> int: ...
    def ReadNextRay(self) -> bool: ...
    def ReadNextRaySegment(self) -> bool: ...
    def ReadNextResult(self) -> tuple[bool, int, int, float, int]: ...
    def ReadNextSegment(
        self,
    ) -> tuple[
        bool,
        int,
        int,
        int,
        int,
        int,
        RayStatus,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]: ...
    def ReadNextSegmentFull(
        self,
    ) -> tuple[
        bool,
        int,
        int,
        int,
        int,
        int,
        RayStatus,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
        int,
        int,
        float,
        float,
        float,
        float,
        float,
        float,
        float,
    ]: ...

class LTEdgeSasmpling:
    S_1X_Low = 0
    S_4X = 1
    S_16X = 2
    S_64X = 3
    S_256X = 4

class LTRaySampling:
    S_1X_Low = 0
    S_4X = 1
    S_16X = 2
    S_64X = 3
    S_256X = 4
    S_1024X = 5

class NSCTraceOptions:
    # None = 0
    UsePolarization = 1
    UseSplitting = 2
    UsePolarizationSplitting = 3
    UseScattering = 4
    UsePolarizationScattering = 5
    UseSplittingScattering = 6
    UsePolarizationSplittingScattering = 7

class OPDMode:
    # None = 0
    Current = 1
    CurrentAndChief = 2

class RayPathDataEntry: ...

class RayPathDataHolder:
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, offset: int, size: int): ...
    @overload
    def __init__(self, info: SerializationInfo, context: StreamingContext): ...
    def GetData(self, mmf: MemoryMappedFile, data: list[RayPathDataEntry]) -> tuple[bool, list[RayPathDataEntry]]: ...
    def GetObjectData(self, info: SerializationInfo, context: StreamingContext) -> None: ...
    def SetData(mmf: MemoryMappedFile, data: list[RayPathDataEntry], index: int) -> RayPathDataHolder: ...

class RayStatus:
    Parent = 0
    Terminated = 1
    Reflected = 2
    Transmitted = 4
    Scattered = 8
    Diffracted = 16
    GhostedFrom = 32
    DiffractedFrom = 64
    ScatteredFrom = 128
    RayError = 256
    BulkScattered = 512
    WaveShifted = 1024
    TIR = 2048
    OrdinaryRay = 65536
    ExtraordinaryRay = 131072
    WaveShiftPL = 262144

class RaysType:
    Real = 0
    Paraxial = 1

class ZRDFormatType:
    UncompressedFullData = 0
    CompressedBasicData = 1
    CompressedFullData = 2



"""Zemax OpticStudio analyses from the Polarization category."""

from __future__ import annotations

import os
import struct
from dataclasses import dataclass
from io import StringIO
from tempfile import mkstemp

import pandas as pd

import zospy.api.config as _config
from zospy import utils
from zospy.analyses.old.base import AnalysisResult, OnComplete, new_analysis
from zospy.api import constants
from zospy.utils.zputils import _get_number_field
from zospy.zpcore import OpticStudioSystem


@dataclass
class PupilMapData:
    """Pupil map analysis data."""

    Header: str
    Wavelength: float
    FieldPos: float
    XField: float
    YField: float
    XPhase: float
    YPhase: float
    Configs: int
    Surface: int
    Transmission: float
    Table: pd.DataFrame


def polarization_pupil_map(
    oss: OpticStudioSystem,
    jx: float = 1,
    jy: float = 0,
    x_phase: float = 0,
    y_phase: float = 0,
    wavelength: int = 1,
    field: int = 1,
    surface: str | int = "Image",
    sampling: str | int = "11x11",
    add_configs: str = "",
    sub_configs: str = "",
    oncomplete: OnComplete | str = OnComplete.Close,
    cfgoutfile: str | None = None,
    txtoutfile: str | None = None,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Polarization Pupil Map Analysis.

    Due to limitations in the ZOS-API, the output is obtained by writing the OpticStudio results to a file and
    subsequently reading in this file

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    jx : float
        Jones x electric field. Defaults to 1.
    jy : float
        Jones y electric field. Defaults to 0.
    x_phase : float
        Phase of the X component of the Jones electric field in degrees. Defaults to 0.
    y_phase : float
        Phase of the Y component of the Jones electric field in degrees. Defaults to 0.
    wavelength : int
        The wavelength number that is to be used. Should be an integer specifying the wavelength number.
        Defaults to 1.
    field : int
        The field number that is to be used. Must be an integer specifying the field number. Defaults
        to 1.
    surface : str or int
        The surface that is to be analyzed. Either 'Image', or an integer. Defaults to 'Image'.
    sampling : str or int
        The size of the used grid, either string (e.g. '65x65') or int. The integer will be treated as if obtained from
        zospy.constants.Analysis.SampleSizes_ContrastLoss. Defaults to '11x11'.
    add_configs : str
        The add configs string.
    sub_configs : str
        The subtract configs string.
    oncomplete : str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.
    cfgoutfile : str or None
        The configuration file to which the current configuration is saved. If None, a temporary file will be created
        and deleted. If string, it should be a full system path with '.CFG' as extension, after which the file will be
        saved in that location. Defaults to
        None.
    txtoutfile : str or None
        The textfile to which the OpticStudio output is saved. If None, a temporary file will be created and deleted. If
        string, it should be a full system path with '.txt' as extension, after which the file will be saved in that
        location. Defaults to None.

    Returns
    -------
    AnalysisResult
        A Polarization Pupil Map. Next to the standard data, the raw text return obtained from the analysis
        will be present under 'RawTextData', and the txtoutfile under 'TxtOutFile'.
    """
    analysis_type = constants.Analysis.AnalysisIDM.PolarizationPupilMap

    if cfgoutfile is None:
        fd, cfgoutfile = mkstemp(suffix=".CFG", prefix="zospy_")
        os.close(fd)
        cleancfg = True
    else:
        if not cfgoutfile.endswith(".CFG"):
            raise ValueError('cfgfile should end with ".CFG"')
        cleancfg = False

    if txtoutfile is None:
        fd, txtoutfile = mkstemp(suffix=".txt", prefix="zospy_")
        os.close(fd)
        cleantxt = True
    else:
        if not txtoutfile.endswith(".txt"):
            raise ValueError('txtfile should end with ".txt"')
        cleantxt = False

    analysis = new_analysis(oss, analysis_type)

    # Modify the settings file
    analysis_settings = analysis.GetSettings()
    analysis_settings.SaveTo(cfgoutfile)

    # MODIFYSETTINGS are defined in the ZPL help files: The Programming Tab > About the ZPL > Keywords
    analysis_settings.ModifySettings(cfgoutfile, "PPM_JX", utils.pyutils.xtoa(jx, thousands_separator=None))
    analysis_settings.ModifySettings(cfgoutfile, "PPM_JY", utils.pyutils.xtoa(jy, thousands_separator=None))
    analysis_settings.ModifySettings(cfgoutfile, "PPM_PX", utils.pyutils.xtoa(x_phase, thousands_separator=None))
    analysis_settings.ModifySettings(cfgoutfile, "PPM_PY", utils.pyutils.xtoa(y_phase, thousands_separator=None))
    analysis_settings.ModifySettings(
        cfgoutfile, "PPM_WAVE", utils.pyutils.xtoa(int(wavelength), thousands_separator=None)
    )
    analysis_settings.ModifySettings(cfgoutfile, "PPM_FIELD", utils.pyutils.xtoa(int(field), thousands_separator=None))
    if isinstance(surface, str):
        analysis_settings.ModifySettings(cfgoutfile, "PPM_SURFACE", surface)
    else:
        analysis_settings.ModifySettings(
            cfgoutfile, "PPM_SURFACE", utils.pyutils.xtoa(surface, thousands_separator=None)
        )
    sampling_value = getattr(
        constants.Analysis.SampleSizes_ContrastLoss,
        utils.zputils.standardize_sampling(sampling),
    ).value__
    analysis_settings.ModifySettings(
        cfgoutfile, "PPM_SAMP", utils.pyutils.xtoa(sampling_value - 1, thousands_separator=None)
    )
    analysis_settings.ModifySettings(cfgoutfile, "PPM_ADDCONFIG", str(add_configs))
    analysis_settings.ModifySettings(cfgoutfile, "PPM_SUBCONFIGS", str(sub_configs))

    analysis_settings.LoadFrom(cfgoutfile)

    # Run analysis
    analysis.ApplyAndWaitForCompletion()

    # Get results
    analysis.Results.GetTextFile(txtoutfile)

    with open(txtoutfile, "r", encoding=oss.ZOS.get_txtfile_encoding()) as f:
        text_output = f.read()
        line_list = text_output.split("\n")

    # Get header
    header = line_list[0].strip().replace("\ufeff", "")

    # Read data table as dataframe
    df = pd.read_csv(
        StringIO("\n".join(line_list[17:]).replace(" ", "")),
        delimiter="\t",
        decimal=_config.DECIMAL_POINT,
    )

    data = PupilMapData(
        Header=header,
        Wavelength=utils.pyutils.atox(_get_number_field("Wavelength", text_output), dtype=float),
        FieldPos=utils.pyutils.atox(_get_number_field("Field Pos", text_output), dtype=float),
        XField=utils.pyutils.atox(_get_number_field("X-Field", text_output), dtype=float),
        YField=utils.pyutils.atox(_get_number_field("Y-Field", text_output), dtype=float),
        XPhase=utils.pyutils.atox(_get_number_field("X-Phase", text_output), dtype=float),
        YPhase=utils.pyutils.atox(_get_number_field("Y-Phase", text_output), dtype=float),
        Configs=utils.pyutils.atox(_get_number_field("Configs", text_output), dtype=int),
        Surface=utils.pyutils.atox(_get_number_field("Surface", text_output), dtype=int),
        Transmission=utils.pyutils.atox(_get_number_field("Transmission", text_output), dtype=float),
        Table=df,
    )

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)

    settings.loc["Jx"] = jx
    settings.loc["Jy"] = jy
    settings.loc["X-Phase"] = x_phase
    settings.loc["Y-Phase"] = y_phase
    settings.loc["Wavelength"] = int(wavelength)
    settings.loc["Field"] = int(field)
    settings.loc["Surface"] = surface
    settings.loc["Sampling"] = sampling
    settings.loc["Add Configs"] = add_configs
    settings.loc["Sub Configs"] = sub_configs

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
        RawTextData=line_list,
        CgfOutFile=cfgoutfile,
        TxtOutFile=txtoutfile,
    )

    # cleanup if needed
    if cleancfg:
        os.remove(cfgoutfile)
    if cleantxt:
        os.remove(txtoutfile)

    return analysis.complete(oncomplete, result)


@dataclass
class TransmissionData:
    """Transmission analysis data."""

    # TODO add support for multiple fields and wavelengths
    Header: str
    FieldPos: float
    TotalTransmission: float
    Wavelength: float
    Table: pd.DataFrame


def transmission(
    oss: OpticStudioSystem,
    sampling: str | int = "32x32",
    unpolarized: bool = False,
    jx: float = 1,
    jy: float = 0,
    x_phase: float = 0,
    y_phase: float = 0,
    oncomplete: OnComplete | str = OnComplete.Close,
    cfgoutfile: str | None = None,
    txtoutfile: str | None = None,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Polarization Transmission Analysis.

    Due to limitations in the ZOS-API, the output is obtained by writing the OpticStudio results to a file and
    subsequently reading in this file.

    WARNING: This analysis currently only works for systems with a single wavelength and field. A
    `NotImplementedException` will be raised if multiple fields or wavelengths are used. See the discussion in
    https://github.com/MREYE-LUMC/ZOSPy/pull/14 for more details.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    sampling : str or int
        The size of the used grid, either string (e.g. '128x128') or int. The integer will be treated as if obtained
        from zospy.constants.Analysis.SampleSizes. Defaults to '32x32'.
    unpolarized : bool
        Defines if unpolarized light is used. Defaults to False.
    jx : float
        Jones x electric field. Defaults to 1.
    jy : float
        Jones y electric field. Defaults to 0.
    x_phase : float
        Phase of the X component of the Jones electric field in degrees. Defaults to 0.
    y_phase : float
        Phase of the Y component of the Jones electric field in degrees. Defaults to 0.
    oncomplete : str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to 'Close'.
    cfgoutfile : str or None
        The configuration file to which the current configuration is saved. If None, a temporary file will be created
        and deleted. If string, it should be a full system path with '.CFG' as extension, after which the file will be
        saved in that location. Defaults to
        None.
    txtoutfile : str or None
        The textfile to which the OpticStudio output is saved. If None, a temporary file will be created and deleted. If
        string, it should be a full system path with '.txt' as extension, after which the file will be saved in that
        location. Defaults to None.

    Returns
    -------
    AnalysisResult
        A Polarization Transmission Analysis. Next to the standard data, the raw text return obtained from the analysis
        will be present under 'RawTextData', and the txtoutfile under 'TxtOutFile'.

    Raises
    ------
    NotImplementedException
        If multiple fields or wavelenghts are present in the optical system/
    """
    if oss.SystemData.Fields.NumberOfFields > 1 or oss.SystemData.Wavelengths.NumberOfWavelengths > 1:
        raise NotImplementedError("Only systems with a single field and a single wavelength are currently supported.")

    analysis_type = constants.Analysis.AnalysisIDM.Transmission

    if cfgoutfile is None:
        fd, cfgoutfile = mkstemp(suffix=".CFG", prefix="zospy_")
        os.close(fd)
        cleancfg = True
    else:
        if not cfgoutfile.endswith(".CFG"):
            raise ValueError('cfgfile should end with ".CFG"')
        cleancfg = False

    if txtoutfile is None:
        fd, txtoutfile = mkstemp(suffix=".txt", prefix="zospy_")
        os.close(fd)
        cleantxt = True
    else:
        if not txtoutfile.endswith(".txt"):
            raise ValueError('txtfile should end with ".txt"')
        cleantxt = False

    analysis = new_analysis(oss, analysis_type)

    # Modify the settings file
    analysis_settings = analysis.GetSettings()
    analysis_settings.SaveTo(cfgoutfile)

    settings_bytestring = b"".join(open(cfgoutfile, "rb").readlines())
    settings_bytearray = bytearray(settings_bytestring)

    # Change settings - all byte indices could only be found via reverse engineering :(
    sampling_value = getattr(constants.Analysis.SampleSizes, utils.zputils.standardize_sampling(sampling)).value__
    settings_bytearray[56] = sampling_value
    settings_bytearray[60] = int(unpolarized)
    settings_bytearray[24:32] = struct.pack("<d", jx)
    settings_bytearray[32:40] = struct.pack("<d", jy)
    settings_bytearray[40:48] = struct.pack("<d", x_phase)
    settings_bytearray[48:56] = struct.pack("<d", y_phase)

    with open(cfgoutfile, "wb") as bfile:
        bfile.write(settings_bytearray)

    analysis_settings.LoadFrom(cfgoutfile)

    # Run analysis
    analysis.ApplyAndWaitForCompletion()

    # Get results
    analysis.Results.GetTextFile(txtoutfile)

    with open(txtoutfile, "r", encoding=oss.ZOS.get_txtfile_encoding()) as f:
        text_output = f.read()
        line_list = text_output.split("\n")

    # Get header
    header = line_list[0].strip().replace("\ufeff", "")

    # Read data table as dataframe
    df = pd.read_csv(
        StringIO("\n".join(line_list[26:]).strip()),
        delimiter="\t",
        decimal=_config.DECIMAL_POINT,
    )
    df.columns = df.columns.str.strip()

    data = TransmissionData(
        Header=header,
        FieldPos=utils.pyutils.atox(_get_number_field("Field Pos", text_output), dtype=float),
        Wavelength=utils.pyutils.atox(_get_number_field("Wavelength 1", text_output), dtype=float),
        TotalTransmission=utils.pyutils.atox(_get_number_field("Total Transmission", text_output), dtype=float),
        Table=df,
    )

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Get settings
    settings = pd.Series(name="Settings", dtype=object)

    settings.loc["Sampling"] = sampling
    settings.loc["Unpolarized"] = unpolarized
    settings.loc["Jx"] = jx
    settings.loc["Jy"] = jy
    settings.loc["X-Phase"] = x_phase
    settings.loc["Y-Phase"] = y_phase

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=data,
        settings=settings,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
        RawTextData=line_list,
        CgfOutFile=cfgoutfile,
        TxtOutFile=txtoutfile,
    )

    # cleanup if needed
    if cleancfg:
        os.remove(cfgoutfile)
    if cleantxt:
        os.remove(txtoutfile)

    return analysis.complete(oncomplete, result)
